import math
from math import pi
import numpy.testing as nt
import unittest

from spatialmath import *
from spatialmath.base import *
from spatialmath.baseposematrix import BasePoseMatrix

import numpy as np
from math import pi, sin, cos


def qcompare(x, y):
    if isinstance(x, Quaternion):
        x = x.vec
    elif isinstance(x, BasePoseMatrix):
        x = x.A
    if isinstance(y, Quaternion):
        y = y.vec
    elif isinstance(y, BasePoseMatrix):
        y = y.A
    nt.assert_array_almost_equal(x, y)

# straight port of the MATLAB unit tests


class TestUnitQuaternion(unittest.TestCase):

    def test_constructor_variants(self):
        nt.assert_array_almost_equal(UnitQuaternion().vec, np.r_[1, 0, 0, 0])

        nt.assert_array_almost_equal(UnitQuaternion.Rx(90, 'deg').vec, np.r_[1, 1, 0, 0] / math.sqrt(2))
        nt.assert_array_almost_equal(UnitQuaternion.Rx(-90, 'deg').vec, np.r_[1, -1, 0, 0] / math.sqrt(2))
        nt.assert_array_almost_equal(UnitQuaternion.Ry(90, 'deg').vec, np.r_[1, 0, 1, 0] / math.sqrt(2))
        nt.assert_array_almost_equal(UnitQuaternion.Ry(-90, 'deg').vec, np.r_[1, 0, -1, 0] / math.sqrt(2))
        nt.assert_array_almost_equal(UnitQuaternion.Rz(90, 'deg').vec, np.r_[1, 0, 0, 1] / math.sqrt(2))
        nt.assert_array_almost_equal(UnitQuaternion.Rz(-90, 'deg').vec, np.r_[1, 0, 0, -1] / math.sqrt(2))


    def test_constructor(self):

        qcompare(UnitQuaternion(), [1, 0, 0, 0])

        # from S
        qcompare(UnitQuaternion([1, 0, 0, 0]), np.r_[1, 0, 0, 0])
        qcompare(UnitQuaternion([0, 1, 0, 0]), np.r_[0, 1, 0, 0])
        qcompare(UnitQuaternion([0, 0, 1, 0]), np.r_[0, 0, 1, 0])
        qcompare(UnitQuaternion([0, 0, 0, 1]), np.r_[0, 0, 0, 1])

        qcompare(UnitQuaternion([2, 0, 0, 0]), np.r_[1, 0, 0, 0])
        qcompare(UnitQuaternion([-2, 0, 0, 0]), np.r_[1, 0, 0, 0])

        # from [S,V]
        qcompare(UnitQuaternion(1, [0, 0, 0]), np.r_[1, 0, 0, 0])
        qcompare(UnitQuaternion(0, [1, 0, 0]), np.r_[0, 1, 0, 0])
        qcompare(UnitQuaternion(0, [0, 1, 0]), np.r_[0, 0, 1, 0])
        qcompare(UnitQuaternion(0, [0, 0, 1]), np.r_[0, 0, 0, 1])

        qcompare(UnitQuaternion(2, [0, 0, 0]), np.r_[1, 0, 0, 0])
        qcompare(UnitQuaternion(-2, [0, 0, 0]), np.r_[1, 0, 0, 0])

        # from R

        qcompare(UnitQuaternion(np.eye(3)), [1, 0, 0, 0])

        qcompare(UnitQuaternion(rotx(pi / 2)), np.r_[1, 1, 0, 0] / math.sqrt(2))
        qcompare(UnitQuaternion(roty(pi / 2)), np.r_[1, 0, 1, 0] / math.sqrt(2))
        qcompare(UnitQuaternion(rotz(pi / 2)), np.r_[1, 0, 0, 1] / math.sqrt(2))

        qcompare(UnitQuaternion(rotx(-pi / 2)), np.r_[1, -1, 0, 0] / math.sqrt(2))
        qcompare(UnitQuaternion(roty(-pi / 2)), np.r_[1, 0, -1, 0] / math.sqrt(2))
        qcompare(UnitQuaternion(rotz(-pi / 2)), np.r_[1, 0, 0, -1] / math.sqrt(2))

        qcompare(UnitQuaternion(rotx(pi)), np.r_[0, 1, 0, 0])
        qcompare(UnitQuaternion(roty(pi)), np.r_[0, 0, 1, 0])
        qcompare(UnitQuaternion(rotz(pi)), np.r_[0, 0, 0, 1])

        # from SO3

        qcompare(UnitQuaternion(SO3()), np.r_[1, 0, 0, 0])

        qcompare(UnitQuaternion(SO3.Rx(pi / 2)), np.r_[1, 1, 0, 0] / math.sqrt(2))
        qcompare(UnitQuaternion(SO3.Ry(pi / 2)), np.r_[1, 0, 1, 0] / math.sqrt(2))
        qcompare(UnitQuaternion(SO3.Rz(pi / 2)), np.r_[1, 0, 0, 1] / math.sqrt(2))

        qcompare(UnitQuaternion(SO3.Rx(-pi / 2)), np.r_[1, -1, 0, 0] / math.sqrt(2))
        qcompare(UnitQuaternion(SO3.Ry(-pi / 2)), np.r_[1, 0, -1, 0] / math.sqrt(2))
        qcompare(UnitQuaternion(SO3.Rz(-pi / 2)), np.r_[1, 0, 0, -1] / math.sqrt(2))

        qcompare(UnitQuaternion(SO3.Rx(pi)), np.r_[0, 1, 0, 0])
        qcompare(UnitQuaternion(SO3.Ry(pi)), np.r_[0, 0, 1, 0])
        qcompare(UnitQuaternion(SO3.Rz(pi)), np.r_[0, 0, 0, 1])

        # vector of SO3
        q = UnitQuaternion([SO3.Rx(pi / 2), SO3.Ry(pi / 2), SO3.Rz(pi / 2)])
        self.assertEqual(len(q), 3)
        qcompare(q, np.array([[1, 1, 0, 0], [1, 0, 1, 0], [1, 0, 0, 1]]) / math.sqrt(2))

        # from SE3

        qcompare(UnitQuaternion(SE3()), np.r_[1, 0, 0, 0])

        qcompare(UnitQuaternion(SE3.Rx(pi / 2)), np.r_[1, 1, 0, 0] / math.sqrt(2))
        qcompare(UnitQuaternion(SE3.Ry(pi / 2)), np.r_[1, 0, 1, 0] / math.sqrt(2))
        qcompare(UnitQuaternion(SE3.Rz(pi / 2)), np.r_[1, 0, 0, 1] / math.sqrt(2))

        qcompare(UnitQuaternion(SE3.Rx(-pi / 2)), np.r_[1, -1, 0, 0] / math.sqrt(2))
        qcompare(UnitQuaternion(SE3.Ry(-pi / 2)), np.r_[1, 0, -1, 0] / math.sqrt(2))
        qcompare(UnitQuaternion(SE3.Rz(-pi / 2)), np.r_[1, 0, 0, -1] / math.sqrt(2))

        qcompare(UnitQuaternion(SE3.Rx(pi)), np.r_[0, 1, 0, 0])
        qcompare(UnitQuaternion(SE3.Ry(pi)), np.r_[0, 0, 1, 0])
        qcompare(UnitQuaternion(SE3.Rz(pi)), np.r_[0, 0, 0, 1])

        # vector of SE3
        q = UnitQuaternion([SE3.Rx(pi / 2), SE3.Ry(pi / 2), SE3.Rz(pi / 2)])
        self.assertEqual(len(q), 3)
        qcompare(q, np.array([[1, 1, 0, 0], [1, 0, 1, 0], [1, 0, 0, 1]]) / math.sqrt(2))


        # from S
        M = np.identity(4)
        q = UnitQuaternion(M)
        self.assertEqual(len(q), 4)

        qcompare(q[0], np.r_[1, 0, 0, 0])
        qcompare(q[1], np.r_[0, 1, 0, 0])
        qcompare(q[2], np.r_[0, 0, 1, 0])
        qcompare(q[3], np.r_[0, 0, 0, 1])


        # # vectorised forms of R, T
        # R = []; T = []
        # for theta in [-pi/2, 0, pi/2, pi]:
        #     R = cat(3, R, rotx(theta), roty(theta), rotz(theta))
        #     T = cat(3, T, trotx(theta), troty(theta), trotz(theta))

        # nt.assert_array_almost_equal(UnitQuaternion(R).R, R)
        # nt.assert_array_almost_equal(UnitQuaternion(T).T, T)

        # copy constructor
        q = UnitQuaternion(rotx(0.3))
        qcompare(UnitQuaternion(q), q)

    def test_concat(self):
        u = UnitQuaternion()
        uu = UnitQuaternion([u, u, u, u])

        self.assertIsInstance(uu, UnitQuaternion)
        self.assertEqual(len(uu), 4)

    def test_string(self):

        u = UnitQuaternion()

        s = str(u)
        self.assertIsInstance(s, str)
        self.assertTrue(s.endswith(' >>'))
        self.assertEqual(s.count('\n'), 0)

        q = UnitQuaternion.Rx([0.3, 0.4, 0.5])
        s = str(q)
        self.assertIsInstance(s, str)
        self.assertEqual(s.count('\n'), 2)

    def test_properties(self):

        u = UnitQuaternion()

        # s,v
        nt.assert_array_almost_equal(UnitQuaternion([1, 0, 0, 0]).s, 1)
        nt.assert_array_almost_equal(UnitQuaternion([1, 0, 0, 0]).v, [0, 0, 0])

        nt.assert_array_almost_equal(UnitQuaternion([0, 1, 0, 0]).s, 0)
        nt.assert_array_almost_equal(UnitQuaternion([0, 1, 0, 0]).v, [1, 0, 0])

        nt.assert_array_almost_equal(UnitQuaternion([0, 0, 1, 0]).s, 0)
        nt.assert_array_almost_equal(UnitQuaternion([0, 0, 1, 0]).v, [0, 1, 0])

        nt.assert_array_almost_equal(UnitQuaternion([0, 0, 0, 1]).s, 0)
        nt.assert_array_almost_equal(UnitQuaternion([0, 0, 0, 1]).v, [0, 0, 1])

        # R,T
        nt.assert_array_almost_equal(u.R, np.eye(3))

        nt.assert_array_almost_equal(UnitQuaternion(rotx(pi / 2)).R, rotx(pi / 2))
        nt.assert_array_almost_equal(UnitQuaternion(roty(-pi / 2)).R, roty(-pi / 2))
        nt.assert_array_almost_equal(UnitQuaternion(rotz(pi)).R, rotz(pi))

        qcompare(UnitQuaternion(rotx(pi / 2)).SO3(), SO3.Rx(pi / 2))
        qcompare(UnitQuaternion(roty(-pi / 2)).SO3(), SO3.Ry(-pi / 2))
        qcompare(UnitQuaternion(rotz(pi)).SO3(), SO3.Rz(pi))

        qcompare(UnitQuaternion(rotx(pi / 2)).SE3(), SE3.Rx(pi / 2))
        qcompare(UnitQuaternion(roty(-pi / 2)).SE3(), SE3.Ry(-pi / 2))
        qcompare(UnitQuaternion(rotz(pi)).SE3(), SE3.Rz(pi))


    def test_staticconstructors(self):
        # rotation primitives
        for theta in [-pi / 2, 0, pi / 2, pi]:
            nt.assert_array_almost_equal(UnitQuaternion.Rx(theta).R, rotx(theta))

        for theta in [-pi / 2, 0, pi / 2, pi]:
            nt.assert_array_almost_equal(UnitQuaternion.Ry(theta).R, roty(theta))

        for theta in [-pi / 2, 0, pi / 2, pi]:
            nt.assert_array_almost_equal(UnitQuaternion.Rz(theta).R, rotz(theta))

        for theta in np.r_[-pi / 2, 0, pi / 2, pi] * 180 / pi:
            nt.assert_array_almost_equal(UnitQuaternion.Rx(theta, 'deg').R, rotx(theta, 'deg'))

        for theta in [-pi / 2, 0, pi / 2, pi]:
            nt.assert_array_almost_equal(UnitQuaternion.Ry(theta, 'deg').R, roty(theta, 'deg'))

        for theta in [-pi / 2, 0, pi / 2, pi]:
            nt.assert_array_almost_equal(UnitQuaternion.Rz(theta, 'deg').R, rotz(theta, 'deg'))

        # 3 angle
        nt.assert_array_almost_equal(UnitQuaternion.RPY([0.1, 0.2, 0.3]).R, rpy2r(0.1, 0.2, 0.3))

        nt.assert_array_almost_equal(UnitQuaternion.Eul([0.1, 0.2, 0.3]).R, eul2r(0.1, 0.2, 0.3))

        nt.assert_array_almost_equal(UnitQuaternion.RPY([10, 20, 30], unit='deg').R, rpy2r(10, 20, 30, unit='deg'))

        nt.assert_array_almost_equal(UnitQuaternion.Eul([10, 20, 30], unit='deg').R, eul2r(10, 20, 30, unit='deg'))

        # (theta, v)
        th = 0.2
        v = unitvec([1, 2, 3])
        nt.assert_array_almost_equal(UnitQuaternion.AngVec(th, v).R, angvec2r(th, v))
        nt.assert_array_almost_equal(UnitQuaternion.AngVec(-th, v).R, angvec2r(-th, v))
        nt.assert_array_almost_equal(UnitQuaternion.AngVec(-th, -v).R, angvec2r(-th, -v))
        nt.assert_array_almost_equal(UnitQuaternion.AngVec(th, -v).R, angvec2r(th, -v))

        # (theta, v)
        th = 0.2
        v = unitvec([1, 2, 3])
        nt.assert_array_almost_equal(UnitQuaternion.EulerVec(th * v).R, angvec2r(th, v))
        nt.assert_array_almost_equal(UnitQuaternion.EulerVec(-th * v).R, angvec2r(-th, v))

    def test_canonic(self):
        R = rotx(0)
        qcompare(UnitQuaternion(R), [1, 0, 0, 0])

        R = rotx(pi / 2)
        qcompare(UnitQuaternion(R), np.r_[cos(pi / 4), sin(pi / 4) * np.r_[1, 0, 0]])
        R = roty(pi / 2)
        qcompare(UnitQuaternion(R), np.r_[cos(pi / 4), sin(pi / 4) * np.r_[0, 1, 0]])
        R = rotz(pi / 2)
        qcompare(UnitQuaternion(R), np.r_[cos(pi / 4), sin(pi / 4) * np.r_[0, 0, 1]])

        R = rotx(-pi / 2)
        qcompare(UnitQuaternion(R), np.r_[cos(pi / 4), sin(pi / 4) * np.r_[-1, 0, 0]])
        R = roty(-pi / 2)
        qcompare(UnitQuaternion(R), np.r_[cos(pi / 4), sin(pi / 4) * np.r_[0, -1, 0]])
        R = rotz(-pi / 2)
        qcompare(UnitQuaternion(R), np.r_[cos(pi / 4), sin(pi / 4) * np.r_[0, 0, -1]])

        R = rotx(pi)
        qcompare(UnitQuaternion(R), np.r_[cos(pi / 2), sin(pi / 2) * np.r_[1, 0, 0]])
        R = roty(pi)
        qcompare(UnitQuaternion(R), np.r_[cos(pi / 2), sin(pi / 2) * np.r_[0, 1, 0]])
        R = rotz(pi)
        qcompare(UnitQuaternion(R), np.r_[cos(pi / 2), sin(pi / 2) * np.r_[0, 0, 1]])

        R = rotx(-pi)
        qcompare(UnitQuaternion(R), np.r_[cos(-pi / 2), sin(-pi / 2) * np.r_[1, 0, 0]])
        R = roty(-pi)
        qcompare(UnitQuaternion(R), np.r_[cos(-pi / 2), sin(-pi / 2) * np.r_[0, 1, 0]])
        R = rotz(-pi)
        qcompare(UnitQuaternion(R), np.r_[cos(-pi / 2), sin(-pi / 2) * np.r_[0, 0, 1]])

    def test_convert(self):
        # test conversion from rotn matrix to u.quaternion and back
        R = rotx(0)
        qcompare(UnitQuaternion(R).R, R)

        R = rotx(pi / 2)
        qcompare(UnitQuaternion(R).R, R)
        R = roty(pi / 2)
        qcompare(UnitQuaternion(R).R, R)
        R = rotz(pi / 2)
        qcompare(UnitQuaternion(R).R, R)

        R = rotx(-pi / 2)
        qcompare(UnitQuaternion(R).R, R)
        R = roty(-pi / 2)
        qcompare(UnitQuaternion(R).R, R)
        R = rotz(-pi / 2)
        qcompare(UnitQuaternion(R).R, R)

        R = rotx(pi)
        qcompare(UnitQuaternion(R).R, R)
        R = roty(pi)
        qcompare(UnitQuaternion(R).R, R)
        R = rotz(pi)
        qcompare(UnitQuaternion(R).R, R)

        R = rotx(-pi)
        qcompare(UnitQuaternion(R).R, R)
        R = roty(-pi)
        qcompare(UnitQuaternion(R).R, R)
        R = rotz(-pi)
        qcompare(UnitQuaternion(R).R, R)

    def test_resulttype(self):

        q = Quaternion([2, 0, 0, 0])
        u = UnitQuaternion()

        self.assertIsInstance(q * q, Quaternion)
        self.assertIsInstance(q * u, Quaternion)
        self.assertIsInstance(u * q, Quaternion)
        self.assertIsInstance(u * u, UnitQuaternion)

        # self.assertIsInstance(u.*u, UnitQuaternion)
        # other combos all fail, test this?

        self.assertIsInstance(u / u, UnitQuaternion)

        self.assertIsInstance(u.conj(), UnitQuaternion)
        self.assertIsInstance(u.inv(), UnitQuaternion)
        self.assertIsInstance(u.unit(), UnitQuaternion)
        self.assertIsInstance(q.unit(), UnitQuaternion)

        self.assertIsInstance(q.conj(), Quaternion)

        self.assertIsInstance(q + q, Quaternion)
        self.assertIsInstance(q - q, Quaternion)

        self.assertIsInstance(u + u, Quaternion)
        self.assertIsInstance(u - u, Quaternion)

        # self.assertIsInstance(q+u, Quaternion)
        # self.assertIsInstance(u+q, Quaternion)

        # self.assertIsInstance(q-u, Quaternion)
        # self.assertIsInstance(u-q, Quaternion)
        # TODO test for ValueError in these cases

        self.assertIsInstance(u.SO3(), SO3)
        self.assertIsInstance(u.SE3(), SE3)

    def test_multiply(self):

        vx = np.r_[1, 0, 0]
        vy = np.r_[0, 1, 0]
        vz = np.r_[0, 0, 1]
        rx = UnitQuaternion.Rx(pi / 2)
        ry = UnitQuaternion.Ry(pi / 2)
        rz = UnitQuaternion.Rz(pi / 2)
        u = UnitQuaternion()

        # quat-quat product
        # scalar x scalar

        qcompare(rx * u, rx)
        qcompare(u * rx, rx)

        # vector x vector
        qcompare(UnitQuaternion([ry, rz, rx]) * UnitQuaternion([rx, ry, rz]), UnitQuaternion([ry * rx, rz * ry, rx * rz]))

        # scalar x vector
        qcompare(ry * UnitQuaternion([rx, ry, rz]), UnitQuaternion([ry * rx, ry * ry, ry * rz]))

        # vector x scalar
        qcompare(UnitQuaternion([rx, ry, rz]) * ry, UnitQuaternion([rx * ry, ry * ry, rz * ry]))

        # quatvector product
        # scalar x scalar

        qcompare(rx * vy, vz)

        # scalar x vector
        nt.assert_array_almost_equal(ry * np.c_[vx, vy, vz], np.c_[-vz, vy, vx])

        # vector x scalar
        nt.assert_array_almost_equal(UnitQuaternion([ry, rz, rx]) * vy, np.c_[vy, -vx, vz])

    def test_matmul(self):

        rx = UnitQuaternion.Rx(pi / 2)
        ry = UnitQuaternion.Ry(pi / 2)
        rz = UnitQuaternion.Rz(pi / 2)

        qcompare(rx @ ry, rx * ry)

        qcompare(UnitQuaternion([ry, rz, rx]) @ UnitQuaternion([rx, ry, rz]), UnitQuaternion([ry * rx, rz * ry, rx * rz]))

    # def multiply_test_normalized(self):

    #     vx = [1, 0, 0]; vy = [0, 1, 0]; vz = [0, 0, 1]
    #     rx = UnitQuaternion.Rx(pi/2)
    #     ry = UnitQuaternion.Ry(pi/2)
    #     rz = UnitQuaternion.Rz(pi/2)
    #     u = UnitQuaternion()

    #     # quat-quat product
    #     # scalar x scalar

    #     nt.assert_array_almost_equal(double(rx.*u), double(rx))
    #     nt.assert_array_almost_equal(double(u.*rx), double(rx))

    #     # shouldn't make that much difference here
    #     nt.assert_array_almost_equal(double(rx.*ry), double(rx*ry))
    #     nt.assert_array_almost_equal(double(rx.*rz), double(rx*rz))

    #     #vector x vector
    #     #nt.assert_array_almost_equal([ry, rz, rx] .* [rx, ry, rz], [ry.*rx, rz.*ry, rx.*rz])

    #     # scalar x vector
    #     #nt.assert_array_almost_equal(ry .* [rx, ry, rz], [ry.*rx, ry.*ry, ry.*rz])

    #     #vector x scalar
    #     #nt.assert_array_almost_equal([rx, ry, rz] .* ry, [rx.*ry, ry.*ry, rz.*ry])

    def test_divide(self):

        rx = UnitQuaternion.Rx(pi / 2)
        ry = UnitQuaternion.Ry(pi / 2)
        rz = UnitQuaternion.Rz(pi / 2)
        u = UnitQuaternion()

        # scalar / scalar
        # implicity tests inv

        qcompare(rx / u, rx)
        qcompare(ry / ry, u)

        #vector /vector
        qcompare(UnitQuaternion([ry, rz, rx]) / UnitQuaternion([rx, ry, rz]), UnitQuaternion([ry / rx, rz / ry, rx / rz]))

        #vector / scalar
        qcompare(UnitQuaternion([rx, ry, rz]) / ry, UnitQuaternion([rx / ry, ry / ry, rz / ry]))

        # scalar /vector
        qcompare(ry / UnitQuaternion([rx, ry, rz]), UnitQuaternion([ry / rx, ry / ry, ry / rz]))

    def test_angle(self):
            # angle between quaternions
        # pure
        v = [5, 6, 7]

    def test_conversions(self):

        # , 3 angle
        qcompare(UnitQuaternion.RPY([0.1, 0.2, 0.3]).rpy(), [0.1, 0.2, 0.3])
        qcompare(UnitQuaternion.RPY(0.1, 0.2, 0.3).rpy(), [0.1, 0.2, 0.3])

        qcompare(UnitQuaternion.Eul([0.1, 0.2, 0.3]).eul(), [0.1, 0.2, 0.3])
        qcompare(UnitQuaternion.Eul(0.1, 0.2, 0.3).eul(), [0.1, 0.2, 0.3])


        qcompare(UnitQuaternion.RPY([10, 20, 30], unit='deg').R, rpy2r(10, 20, 30, unit='deg'))

        qcompare(UnitQuaternion.Eul([10, 20, 30], unit='deg').R, eul2r(10, 20, 30, unit='deg'))

        # (theta, v)
        th = 0.2
        v = unitvec([1, 2, 3])
        [a, b] = UnitQuaternion.AngVec(th, v).angvec()
        self.assertAlmostEqual(a, th)
        nt.assert_array_almost_equal(b, v)

        [a, b] = UnitQuaternion.AngVec(-th, v).angvec()
        self.assertAlmostEqual(a, th)
        nt.assert_array_almost_equal(b, -v)

        # null rotation case
        th = 0
        v = unitvec([1, 2, 3])
        [a, b] = UnitQuaternion.AngVec(th, v).angvec()
        self.assertAlmostEqual(a, th)

    #  SO3                     convert to SO3 class
    #  SE3                     convert to SE3 class

    def test_miscellany(self):

        # AbsTol not used since Quaternion supports eq() operator

        rx = UnitQuaternion.Rx(pi / 2)
        ry = UnitQuaternion.Ry(pi / 2)
        rz = UnitQuaternion.Rz(pi / 2)
        u = UnitQuaternion()

        # norm
        qcompare(rx.norm(), 1)
        qcompare(UnitQuaternion([rx, ry, rz]).norm(), [1, 1, 1])

        # unit
        qcompare(rx.unit(), rx)
        qcompare(UnitQuaternion([rx, ry, rz]).unit(), UnitQuaternion([rx, ry, rz]))

        # inner
        nt.assert_array_almost_equal(u.inner(u), 1)
        nt.assert_array_almost_equal(rx.inner(ry), 0.5)
        nt.assert_array_almost_equal(rz.inner(rz), 1)

        q = rx * ry * rz

        qcompare(q**0, u)
        qcompare(q**(-1), q.inv())
        qcompare(q**2, q * q)

        # angle
        # self.assertEqual(angle(u, u), 0)
        # self.assertEqual(angle(u, rx), pi/4)
        # self.assertEqual(angle(u, [rx, u]), pi/4*np.r_[1, 0])
        # self.assertEqual(angle([rx, u], u), pi/4*np.r_[1, 0])
        # self.assertEqual(angle([rx, u], [u, rx]), pi/4*np.r_[1, 1])
        # TODO angle

        # increment
        # w = [0.02, 0.03, 0.04]

        # nt.assert_array_almost_equal(rx.increment(w), rx*UnitQuaternion.omega(w))

    def test_interp(self):

        rx = UnitQuaternion.Rx(pi / 2)
        ry = UnitQuaternion.Ry(pi / 2)
        rz = UnitQuaternion.Rz(pi / 2)
        u = UnitQuaternion()

        q = UnitQuaternion.RPY([.2, .3, .4])

        # from null
        qcompare(q.interp1(0), u)
        qcompare(q.interp1(1), q)

        #self.assertEqual(length(q.interp(linspace(0,1, 10))), 10)
        #self.assertTrue(all( q.interp([0, 1]) == [u, q]))
        # TODO vectorizing

        q0_5 = q.interp1(0.5)
        qcompare(q0_5 * q0_5, q)

        qq = rx.interp1(11)
        self.assertEqual(len(qq), 11)

        # between two quaternions
        qcompare(q.interp(rx, 0), q)
        qcompare(q.interp(rx, 1), rx)

        # test vectorised results
        qq = q.interp(rx, [0, 1])
        self.assertEqual(len(qq), 2)
        qcompare(qq[0], q)
        qcompare(qq[1], rx)

        qq = rx.interp(q, 11)
        self.assertEqual(len(qq), 11)

        #self.assertTrue(all( q.interp([0, 1], dest=rx, ) == [q, rx]))

        # test shortest option
        # q1 = UnitQuaternion.Rx(0.9*pi)
        # q2 = UnitQuaternion.Rx(-0.9*pi)
        # qq = q1.interp(q2, 11)
        # qcompare( qq(6), UnitQuaternion.Rx(0) )
        # qq = q1.interp(q2, 11, 'shortest')
        # qcompare( qq(6), UnitQuaternion.Rx(pi) )
        # TODO interp

    def test_increment(self):
        q = UnitQuaternion()

        q.increment([0, 0, 0])
        qcompare(q, UnitQuaternion())

        q.increment([0, 0, 0], normalize=True)
        qcompare(q, UnitQuaternion())

        for i in range(10):
            q.increment([0.1, 0, 0])
        qcompare(q, UnitQuaternion.Rx(1))

        q = UnitQuaternion()
        for i in range(10):
            q.increment([0.1, 0, 0], normalize=True)
        qcompare(q, UnitQuaternion.Rx(1))


    def test_eq(self):
        q1 = UnitQuaternion([0, 1, 0, 0])
        q2 = UnitQuaternion([0, -1, 0, 0])
        q3 = UnitQuaternion.Rz(pi / 2)

        self.assertTrue(q1 == q1)
        self.assertTrue(q2 == q2)
        self.assertTrue(q3 == q3)
        self.assertTrue(q1 == q2)  # because of double wrapping
        self.assertFalse(q1 == q3)

        nt.assert_array_almost_equal(UnitQuaternion([q1, q1, q1]) == UnitQuaternion([q1, q1, q1]), [True, True, True])
        nt.assert_array_almost_equal(UnitQuaternion([q1, q2, q3]) == UnitQuaternion([q1, q2, q3]), [True, True, True])
        nt.assert_array_almost_equal(UnitQuaternion([q1, q1, q3]) == q1, [True, True, False])
        nt.assert_array_almost_equal(q3 == UnitQuaternion([q1, q1, q3]), [False, False, True])

    def test_logical(self):
        rx = UnitQuaternion.Rx(pi / 2)
        ry = UnitQuaternion.Ry(pi / 2)

        # equality tests
        self.assertTrue(rx == rx)
        self.assertFalse(rx != rx)
        self.assertFalse(rx == ry)

    def test_dot(self):
        q = UnitQuaternion()
        omega = np.r_[1, 2, 3]

        nt.assert_array_almost_equal(q.dot(omega), np.r_[0, omega / 2])
        nt.assert_array_almost_equal(q.dotb(omega), np.r_[0, omega / 2])

        q = UnitQuaternion.Rx(pi / 2)
        qcompare(q.dot(omega), 0.5 * Quaternion.Pure(omega) * q)
        qcompare(q.dotb(omega), 0.5 * q * Quaternion.Pure(omega))

    def test_matrix(self):

        q1 = UnitQuaternion.RPY([0.1, 0.2, 0.3])
        q2 = UnitQuaternion.RPY([0.2, 0.3, 0.4])

        qcompare(q1 * q2, q1.matrix @ q2.vec)

    def test_vec3(self):

        q1 = UnitQuaternion.RPY([0.1, 0.2, 0.3])
        q2 = UnitQuaternion.RPY([0.2, 0.3, 0.4])

        q12 = q1 * q2

        q1v = q1.vec3
        q2v = q2.vec3

        q12v = UnitQuaternion.qvmul(q1v, q2v)

        q12_ = UnitQuaternion.Vec3(q12v)

        qcompare(q12, q12_)

    # def test_display(self):
    #     ry = UnitQuaternion.Ry(pi/2)

    #     ry.plot()
    #     h = ry.plot()
    #     ry.animate()
    #     ry.animate('rgb')
    #     ry.animate( UnitQuaternion.Rx(pi/2), 'rgb' )


class TestQuaternion(unittest.TestCase):

    def test_constructor(self):

        q = Quaternion()
        self.assertEqual(len(q), 1)
        self.assertIsInstance(q, Quaternion)

        nt.assert_array_almost_equal(Quaternion().vec, [0, 0, 0, 0])

        # from S
        nt.assert_array_almost_equal(Quaternion([1, 0, 0, 0]).vec, [1, 0, 0, 0])
        nt.assert_array_almost_equal(Quaternion([0, 1, 0, 0]).vec, [0, 1, 0, 0])
        nt.assert_array_almost_equal(Quaternion([0, 0, 1, 0]).vec, [0, 0, 1, 0])
        nt.assert_array_almost_equal(Quaternion([0, 0, 0, 1]).vec, [0, 0, 0, 1])

        nt.assert_array_almost_equal(Quaternion([2, 0, 0, 0]).vec, [2, 0, 0, 0])
        nt.assert_array_almost_equal(Quaternion([-2, 0, 0, 0]).vec, [-2, 0, 0, 0])

        # from [S,V]
        nt.assert_array_almost_equal(Quaternion(1, [0, 0, 0]).vec, [1, 0, 0, 0])
        nt.assert_array_almost_equal(Quaternion(0, [1, 0, 0]).vec, [0, 1, 0, 0])
        nt.assert_array_almost_equal(Quaternion(0, [0, 1, 0]).vec, [0, 0, 1, 0])
        nt.assert_array_almost_equal(Quaternion(0, [0, 0, 1]).vec, [0, 0, 0, 1])

        nt.assert_array_almost_equal(Quaternion(2, [0, 0, 0]).vec, [2, 0, 0, 0])
        nt.assert_array_almost_equal(Quaternion(-2, [0, 0, 0]).vec, [-2, 0, 0, 0])

        # pure
        v = [5, 6, 7]
        nt.assert_array_almost_equal(Quaternion.Pure(v).vec, [0, ] + v)

        # tc.verifyError( @() Quaternion.pure([1, 2]), 'SMTB:Quaternion:badarg')

        # copy constructor
        q = Quaternion([1, 2, 3, 4])
        nt.assert_array_almost_equal(Quaternion(q).vec, q.vec)

        # errors

        # tc.verifyError( @() Quaternion(2), 'SMTB:Quaternion:badarg')
        # tc.verifyError( @() Quaternion([1, 2, 3]), 'SMTB:Quaternion:badarg')

    def test_string(self):

        u = Quaternion()

        s = str(u)
        self.assertIsInstance(s, str)
        self.assertTrue(s.endswith(' >'))
        self.assertEqual(s.count('\n'), 0)
        self.assertEqual(len(s), 37)

        q = Quaternion([u, u, u])
        s = str(q)
        self.assertIsInstance(s, str)
        self.assertEqual(s.count('\n'), 2)

    def test_properties(self):

        q = Quaternion([1, 2, 3, 4])
        self.assertEqual(q.s, 1)
        nt.assert_array_almost_equal(q.v, np.r_[2, 3, 4])
        nt.assert_array_almost_equal(q.vec, np.r_[1, 2, 3, 4])

    def log_test_exp(self):

        q1 = Quaternion([4, 3, 2, 1])
        q2 = Quaternion([-1, 2, -3, 4])

        nt.assert_array_almost_equal(exp(log(q1)), q1)
        nt.assert_array_almost_equal(exp(log(q2)), q2)

        #nt.assert_array_almost_equal(log(exp(q1)), q1)
        #nt.assert_array_almost_equal(log(exp(q2)), q2)



    def test_concat(self):
        u = Quaternion()
        uu = Quaternion([u, u, u, u])

        self.assertIsInstance(uu, Quaternion)
        self.assertEqual(len(uu), 4)

    def primitive_test_convert(self):

        # s,v
        nt.assert_array_almost_equal(Quaternion([1, 0, 0, 0]).s, 1)
        nt.assert_array_almost_equal(Quaternion([1, 0, 0, 0]).v, [0, 0, 0])

        nt.assert_array_almost_equal(Quaternion([0, 1, 0, 0]).s, 0)
        nt.assert_array_almost_equal(Quaternion([0, 1, 0, 0]).v, [1, 0, 0])

        nt.assert_array_almost_equal(Quaternion([0, 0, 1, 0]).s, 0)
        nt.assert_array_almost_equal(Quaternion([0, 0, 1, 0]).v, [0, 1, 0])

        nt.assert_array_almost_equal(Quaternion([0, 0, 0, 1]).s, 0)
        nt.assert_array_almost_equal(Quaternion([0, 0, 0, 1]).v, [0, 0, 1])

    def test_resulttype(self):

        q = Quaternion([2, 0, 0, 0])

        self.assertIsInstance(q, Quaternion)

        # other combos all fail, test this?

        self.assertIsInstance(q.conj(), Quaternion)
        self.assertIsInstance(q.unit(), UnitQuaternion)

        self.assertIsInstance(q + q, Quaternion)
        self.assertIsInstance(q + q, Quaternion)

    def test_multiply(self):

        q1 = Quaternion([1, 2, 3, 4])
        q2 = Quaternion([4, 3, 2, 1])
        q3 = Quaternion([-1, 2, -3, 4])

        u = Quaternion([1, 0, 0, 0])

        # quat-quat product
        # scalar x scalar

        qcompare(q1 * u, q1)
        qcompare(u * q1, q1)
        qcompare(q1 * q2, [-12, 6, 24, 12])

        q = q1
        q *= q2
        qcompare(q, [-12, 6, 24, 12])

        # vector x vector
        qcompare(Quaternion([q1, u, q2, u, q3, u]) * Quaternion([u, q1, u, q2, u, q3]), Quaternion([q1, q1, q2, q2, q3, q3]))

        q = Quaternion([q1, u, q2, u, q3, u])
        q *= Quaternion([u, q1, u, q2, u, q3])
        qcompare(q, Quaternion([q1, q1, q2, q2, q3, q3]))

        # scalar x vector
        qcompare(q1 * Quaternion([q1, q2, q3]), Quaternion([q1 * q1, q1 * q2, q1 * q3]))

        # vector x scalar
        qcompare(Quaternion([q1, q2, q3]) * q2, Quaternion([q1 * q2, q2 * q2, q3 * q2]))

        # quat-real product
        # scalar x scalar

        v1 = q1.vec
        qcompare(q1 * 5, v1 * 5)
        qcompare(6 * q1, v1 * 6)
        qcompare(-2 * q1, -2 * v1)

        # scalar x vector
        qcompare(5 * Quaternion([q1, q2, q3]), Quaternion([5 * q1, 5 * q2, 5 * q3]))

        # vector x scalar
        qcompare(Quaternion([q1, q2, q3]) * 5, Quaternion([5 * q1, 5 * q2, 5 * q3]))

        # matrix form of multiplication
        qcompare(q1.matrix @ q2.vec, q1 * q2)

        # quat-scalar product
        qcompare(q1 * 2, q1.vec * 2)
        qcompare(Quaternion([q1 * 2, q2 * 2]), Quaternion([q1, q2]) * 2)

        # errors

        # tc.verifyError( @() q1 * [1, 2, 3], 'SMTB:Quaternion:badarg')
        # tc.verifyError( @() [1, 2, 3]*q1, 'SMTB:Quaternion:badarg')
        # tc.verifyError( @() [q1, q1] * [q1, q1, q1], 'SMTB:Quaternion:badarg')
        # tc.verifyError( @() q1*SE3, 'SMTB:Quaternion:badarg')

    def test_equality(self):
        q1 = Quaternion([1, 2, 3, 4])
        q2 = Quaternion([-2, 1, -4, 3])

        self.assertTrue(q1 == q1)
        self.assertFalse(q1 == q2)

        self.assertTrue(q1 != q2)
        self.assertFalse(q2 != q2)

        qt1 = Quaternion([q1, q1, q2, q2])
        qt2 = Quaternion([q1, q2, q2, q1])

        self.assertEqual(qt1 == q1, [True, True, False, False])
        self.assertEqual(q1 == qt1, [True, True, False, False])
        self.assertEqual(qt1 == qt1, [True, True, True, True])

        self.assertEqual(qt2 == q1, [True, False, False, True])
        self.assertEqual(q1 == qt2, [True, False, False, True])
        self.assertEqual(qt1 == qt2, [True, False, True, False])

        self.assertEqual(qt1 != q1, [False, False, True, True])
        self.assertEqual(q1 != qt1, [False, False, True, True])
        self.assertEqual(qt1 != qt1, [False, False, False, False])

        self.assertEqual(qt2 != q1, [False, True, True, False])
        self.assertEqual(q1 != qt2, [False, True, True, False])
        self.assertEqual(qt1 != qt2, [False, True, False, True])

        # errors

        # tc.verifyError( @() [q1 q1] == [q1 q1 q1], 'SMTB:Quaternion:badarg')
        # tc.verifyError( @() [q1 q1] != [q1 q1 q1], 'SMTB:Quaternion:badarg')

    def basic_test_multiply(self):
        # test run multiplication tests on quaternions
        q = Quaternion([1, 0, 0, 0]) * Quaternion([1, 0, 0, 0])
        qcompare(q.vec, [1, 0, 0, 0])

        q = Quaternion([1, 0, 0, 0]) * Quaternion([1, 2, 3, 4])
        qcompare(q.vec, [1, 2, 3, 4])

        q = Quaternion([1, 2, 3, 4]) * Quaternion([1, 2, 3, 4])
        qcompare(q.vec, [-28, 4, 6, 8])

    def add_test_sub(self):
        v1 = [1, 2, 3, 4]
        v2 = [2, 2, 4, 7]

        # plus
        q = Quaternion(v1) + Quaternion(v2)
        q2 = Quaternion(v1) + v2

        qcompare(q.vec, v1 + v2)
        qcompare(q2.vec, v1 + v2)

        # minus
        q = Quaternion(v1) - Quaternion(v2)
        q2 = Quaternion(v1) - v2
        qcompare(q.vec, v1 - v2)
        qcompare(q2.vec, v1 - v2)

    def test_power(self):

        q = Quaternion([1, 2, 3, 4])

        qcompare(q**0, Quaternion([1, 0, 0, 0]))
        qcompare(q**1, q)
        qcompare(q**2, q * q)

    def test_miscellany(self):
        v = np.r_[1, 2, 3, 4]
        q = Quaternion(v)
        u = Quaternion([1, 0, 0, 0])

        # norm
        nt.assert_array_almost_equal(q.norm(), np.linalg.norm(v))
        nt.assert_array_almost_equal(Quaternion([q, u, q]).norm(), [np.linalg.norm(v), 1, np.linalg.norm(v)])

        # unit
        qu = q.unit()
        uu = UnitQuaternion()
        self.assertIsInstance(q, Quaternion)
        nt.assert_array_almost_equal(qu.vec, v / np.linalg.norm(v))
        qcompare(Quaternion([q, u, q]).unit(), UnitQuaternion([qu, uu, qu]))

        # inner
        nt.assert_equal(u.inner(u), 1)
        nt.assert_equal(q.inner(q), q.norm()**2)
        nt.assert_equal(q.inner(u), np.dot(q.vec, u.vec))


# ---------------------------------------------------------------------------------------#
if __name__ == '__main__':

    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Jul  5 14:37:24 2020

@author: corkep
"""

from spatialmath.geom3d import *

import unittest
import numpy.testing as nt
import spatialmath.base as base

class Line3Test(unittest.TestCase):
    

    # Primitives
    def test_constructor1(self):
        
        # construct from 6-vector
        L = Line3([1, 2, 3, 4, 5, 6])
        self.assertIsInstance(L, Line3)
        nt.assert_array_almost_equal(L.v, np.r_[1, 2, 3])
        nt.assert_array_almost_equal(L.w, np.r_[4, 5, 6])
        
        # construct from object
        L2 = Line3(L)
        self.assertIsInstance(L, Line3)
        nt.assert_array_almost_equal(L2.v, np.r_[1, 2, 3])
        nt.assert_array_almost_equal(L2.w, np.r_[4, 5, 6])
        
        # construct from point and direction
        L = Line3.PointDir([1, 2, 3], [4, 5, 6])
        self.assertTrue(L.contains([1, 2, 3]))
        nt.assert_array_almost_equal(L.uw, base.unitvec([4, 5, 6]))
    
    
    def test_vec(self):
        # verify double
        L = Line3([1, 2, 3, 4, 5, 6])
        nt.assert_array_almost_equal(L.vec, np.r_[1, 2, 3, 4, 5, 6])
    
    def test_constructor2(self):
        # 2, point constructor
        P = np.r_[2, 3, 7]
        Q = np.r_[2, 1, 0]
        L = Line3.Join(P, Q)
        nt.assert_array_almost_equal(L.w, P-Q)
        nt.assert_array_almost_equal(L.v, np.cross(P-Q, Q))
    
        # TODO, all combos of list and ndarray
        # test all possible input shapes
        # L2, = Line3(P, Q)
        # self.assertEqual(double(L2), double(L))
        # L2, = Line3(P, Q')
        # self.assertEqual(double(L2), double(L))
        # L2, = Line3(P', Q')
        # self.assertEqual(double(L2), double(L))
        # L2, = Line3(P, Q)
        # self.assertEqual(double(L2), double(L))
        
        # # planes constructor
        # P = [10, 11, 12]'; w = [1, 2, 3]
        # L = Line3.PointDir(P, w)
        # self.assertEqual(double(L), [cross(w,P) w]'); %FAIL
        # L2, = Line3.PointDir(P', w)
        # self.assertEqual(double(L2), double(L))
        # L2, = Line3.PointDir(P, w')
        # self.assertEqual(double(L2), double(L))
        # L2, = Line3.PointDir(P', w')
        # self.assertEqual(double(L2), double(L))
    
    
    def test_pp(self):
        # validate pp and ppd
        L = Line3.Join([-1, 1, 2], [1, 1, 2])
        nt.assert_array_almost_equal(L.pp, np.r_[0, 1, 2])
        self.assertEqual(L.ppd, math.sqrt(5))
        
        # validate pp
        self.assertTrue( L.contains(L.pp) )
    
    
    def test_contains(self):
        P = [2, 3, 7]
        Q = [2, 1, 0]
        L = Line3.Join(P, Q)
        
        # validate contains
        self.assertTrue( L.contains([2, 3, 7]) )
        self.assertTrue( L.contains([2, 1, 0]) )
        self.assertFalse( L.contains([2, 1, 4]) )
    
    
    def test_closest(self):
        P = [2, 3, 7]
        Q = [2, 1, 0]
        L = Line3.Join(P, Q)
        
        p, d = L.closest_to_point(P)
        nt.assert_array_almost_equal(p, P)
        self.assertAlmostEqual(d, 0)
        
            # validate closest with given points and origin
        p, d = L.closest_to_point(Q)
        nt.assert_array_almost_equal(p, Q)
        self.assertAlmostEqual(d, 0)
        
        L = Line3.Join([-1, 1, 2], [1, 1, 2])
        p, d = L.closest_to_point([0, 1, 2])
        nt.assert_array_almost_equal(p, np.r_[0, 1, 2])
        self.assertAlmostEqual(d, 0)
        
        p, d = L.closest_to_point([5, 1, 2])
        nt.assert_array_almost_equal(p, np.r_[5, 1, 2])
        self.assertAlmostEqual(d, 0)
        
        p, d = L.closest_to_point([0, 0, 0])
        nt.assert_array_almost_equal(p, L.pp)
        self.assertEqual(d, L.ppd)
        
        p, d = L.closest_to_point([5, 1, 0])
        nt.assert_array_almost_equal(p, [5, 1, 2])
        self.assertAlmostEqual(d, 2)
    
    def test_plot(self):
        
        P = [2, 3, 7]
        Q = [2, 1, 0]
        L = Line3.Join(P, Q)
        
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d', proj_type='ortho')
        ax.set_xlim3d(-10, 10)
        ax.set_ylim3d(-10, 10)
        ax.set_zlim3d(-10, 10)
        
        L.plot(color='red', linewidth=2)
    
    def test_eq(self):
        w = np.r_[1, 2, 3]
        P = np.r_[-2, 4, 3]
        
        L1 = Line3.Join(P, P + w)
        L2 = Line3.Join(P + 2 * w, P + 5 * w)
        L3 = Line3.Join(P + np.r_[1, 0, 0], P + w)
        
        self.assertTrue(L1 == L2)
        self.assertFalse(L1 == L3)
        
        self.assertFalse(L1 != L2)
        self.assertTrue(L1 != L3)
    
    def test_skew(self):
        
        P = [2, 3, 7]; Q = [2, 1, 0]
        L = Line3.Join(P, Q)
        
        m = L.skew()
        
        self.assertEqual(m.shape, (4,4))
        nt.assert_array_almost_equal(m + m.T, np.zeros((4,4)))
    
    def test_mtimes(self):
        P = [1, 2, 0]
        Q = [1, 2, 10]  # vertical line through (1,2)
        L = Line3.Join(P, Q)
        
        # check transformation by SE3
        
        L2 = SE3() * L
        nt.assert_array_almost_equal(L.vec, L2.vec)
        
        L2 = SE3(2, 0, 0) * L  # shift line in the x direction
        nt.assert_array_almost_equal(L2.vec, np.r_[20, -30, 0, 0, 0, -10])
        L2 = SE3(0, 2, 0) * L  # shift line in the y direction
        nt.assert_array_almost_equal(L2.vec, np.r_[40, -10, 0, 0, 0, -10])
    
    def test_parallel(self):
        
        L1 = Line3.PointDir([4, 5, 6], [1, 2, 3])
        L2 = Line3.PointDir([5, 5, 6], [1, 2, 3])
        L3 = Line3.PointDir([4, 5, 6], [3, 2, 1])
        
        # L1, || L2, but doesnt intersect
        # L1, intersects L3
        
        self.assertTrue( L1.isparallel(L1) )
        self.assertTrue(L1 | L1)
        
        self.assertTrue( L1.isparallel(L2) )
        self.assertTrue(L1 | L2)
        self.assertTrue( L2.isparallel(L1) )
        self.assertTrue(L2 | L1)
        self.assertFalse( L1.isparallel(L3) )
        self.assertFalse(L1 | L3)
    
    
    def test_intersect(self):
    
        
        L1 = Line3.PointDir([4, 5, 6], [1, 2, 3])
        L2 = Line3.PointDir([5, 5, 6], [1, 2, 3])
        L3 = Line3.PointDir( [4, 5, 6], [0, 0, 1])
        L4 = Line3.PointDir([5, 5, 6], [1, 0, 0])
    
        # L1, || L2, but doesnt intersect
        # L3, intersects L4
        self.assertFalse( L1^L2, )
        
        self.assertTrue( L3^L4, )
        
        
    def test_commonperp(self):
        L1 = Line3.PointDir([4, 5, 6], [0, 0, 1])
        L2 = Line3.PointDir([6, 5, 6], [0, 1, 0])
        
        self.assertFalse( L1|L2)
        self.assertFalse( L1^L2)
        
        self.assertEqual( L1.distance(L2), 2)
        
        L = L1.commonperp(L2)  # common perp intersects both lines
        
        self.assertTrue( L^L1)
        self.assertTrue( L^L2)
    
    
    def test_line(self):
        
        # mindist
        # intersect
        # char
        # intersect_volume
        # mindist
        # mtimes
        # or
        # side
        pass
    
    def test_contains(self):
        P = [2, 3, 7]
        Q = [2, 1, 0]
        L = Line3.Join(P, Q)
        
        self.assertTrue( L.contains(L.point(0)) )
        self.assertTrue( L.contains(L.point(1)) )
        self.assertTrue( L.contains(L.point(-1)) )

    def test_point(self):
        P = [2, 3, 7]
        Q = [2, 1, 0]
        L = Line3.Join(P, Q)
        
        nt.assert_array_almost_equal(L.point(0).flatten(), L.pp)

        for x in (-2, 0, 3):
            nt.assert_array_almost_equal(L.lam(L.point(x)), x)
    
    def test_char(self):
        P = [2, 3, 7]
        Q = [2, 1, 0]
        L = Line3.Join(P, Q)
        
        s = str(L)
        self.assertIsInstance(s, str)


    def test_plane(self):
        
        xyplane = [0, 0, 1, 0]
        xzplane = [0, 1, 0, 0]
        L = Line3.IntersectingPlanes(xyplane, xzplane) # x axis
        nt.assert_array_almost_equal(L.vec, np.r_[0, 0, 0, -1, 0, 0])
        
        L = Line3.Join([-1, 2, 3], [1, 2, 3]);  # line at y=2,z=3
        x6 = [1, 0, 0, -6]  # x = 6
        
        # plane_intersect
        p, lam = L.intersect_plane(x6)
        nt.assert_array_almost_equal(p, np.r_[6, 2, 3])
        nt.assert_array_almost_equal(L.point(lam).flatten(), np.r_[6, 2, 3])
        

        x6s = Plane3.PointNormal(n=[1, 0, 0], p=[6, 0, 0])
        p, lam = L.intersect_plane(x6s)
        nt.assert_array_almost_equal(p, np.r_[6, 2, 3])
        
        nt.assert_array_almost_equal(L.point(lam).flatten(), np.r_[6, 2, 3])
    
    def test_methods(self):
        # intersection
        px = Line3.Join([0, 0, 0], [1, 0, 0]);  # x-axis
        py = Line3.Join([0, 0, 0], [0, 1, 0]);  # y-axis
        px1 = Line3.Join([0, 1, 0], [1, 1, 0]); # offset x-axis
        
        self.assertEqual(px.ppd, 0)
        self.assertEqual(px1.ppd, 1)
        nt.assert_array_almost_equal(px1.pp, [0, 1, 0])

        px.intersects(px)
        px.intersects(py)
        px.intersects(px1)
        
        
    # def test_intersect(self):
    #     px = Line3([0, 0, 0], [1, 0, 0]);  # x-axis
    #     py = Line3([0, 0, 0], [0, 1, 0]);  # y-axis
    #     
    #     plane.d = [1, 0, 0]; plane.p = 2; # plane x=2
    #     
    #     px.intersect_plane(plane)
    #     py.intersect_plane(plane)

if __name__ == "__main__":

    unittest.main()




import unittest
import numpy.testing as nt
import numpy as np

from spatialmath.spatialvector import *


class TestSpatialVector(unittest.TestCase):
    def test_list_powers(self):
        x = SpatialVelocity.Empty()
        self.assertEqual(len(x), 0)
        x.append(SpatialVelocity([1, 2, 3, 4, 5, 6]))
        self.assertEqual(len(x), 1)

        x.append(SpatialVelocity([7, 8, 9, 10, 11, 12]))
        self.assertEqual(len(x), 2)

        y = x[0]
        self.assertIsInstance(y, SpatialVelocity)
        self.assertEqual(len(y), 1)
        self.assertTrue(all(y.A == np.r_[1, 2, 3, 4, 5, 6]))

        y = x[1]
        self.assertIsInstance(y, SpatialVelocity)
        self.assertEqual(len(y), 1)
        self.assertTrue(all(y.A == np.r_[7, 8, 9, 10, 11, 12]))

        x.insert(0, SpatialVelocity([20, 21, 22, 23, 24, 25]))

        y = x[0]
        self.assertIsInstance(y, SpatialVelocity)
        self.assertEqual(len(y), 1)
        self.assertTrue(all(y.A == np.r_[20, 21, 22, 23, 24, 25]))

        y = x[1]
        self.assertIsInstance(y, SpatialVelocity)
        self.assertEqual(len(y), 1)
        self.assertTrue(all(y.A == np.r_[1, 2, 3, 4, 5, 6]))

    def test_velocity(self):
        a = SpatialVelocity([1, 2, 3, 4, 5, 6])
        self.assertIsInstance(a, SpatialVelocity)
        self.assertIsInstance(a, SpatialVector)
        self.assertIsInstance(a, SpatialM6)
        self.assertEqual(len(a), 1)
        self.assertTrue(all(a.A == np.r_[1, 2, 3, 4, 5, 6]))

        a = SpatialVelocity(np.r_[1, 2, 3, 4, 5, 6])
        self.assertIsInstance(a, SpatialVelocity)
        self.assertIsInstance(a, SpatialVector)
        self.assertIsInstance(a, SpatialM6)
        self.assertEqual(len(a), 1)
        self.assertTrue(all(a.A == np.r_[1, 2, 3, 4, 5, 6]))

        s = str(a)
        self.assertIsInstance(s, str)
        self.assertEqual(s.count('\n'), 0)
        self.assertTrue(s.startswith('SpatialVelocity'))

        r = np.random.rand(6, 10)
        a = SpatialVelocity(r)
        self.assertIsInstance(a, SpatialVelocity)
        self.assertIsInstance(a, SpatialVector)
        self.assertIsInstance(a, SpatialM6)
        self.assertEqual(len(a), 10)

        b = a[3]
        self.assertIsInstance(b, SpatialVelocity)
        self.assertIsInstance(b, SpatialVector)
        self.assertIsInstance(b, SpatialM6)
        self.assertEqual(len(b), 1)
        self.assertTrue(all(b.A == r[:,3]))

        s = str(a)
        self.assertIsInstance(s, str)
        self.assertEqual(s.count('\n'), 9)

    def test_acceleration(self):
        a = SpatialAcceleration([1, 2, 3, 4, 5, 6])
        self.assertIsInstance(a, SpatialAcceleration)
        self.assertIsInstance(a, SpatialVector)
        self.assertIsInstance(a, SpatialM6)
        self.assertEqual(len(a), 1)
        self.assertTrue(all(a.A == np.r_[1, 2, 3, 4, 5, 6]))

        a = SpatialAcceleration(np.r_[1, 2, 3, 4, 5, 6])
        self.assertIsInstance(a, SpatialAcceleration)
        self.assertIsInstance(a, SpatialVector)
        self.assertIsInstance(a, SpatialM6)
        self.assertEqual(len(a), 1)
        self.assertTrue(all(a.A == np.r_[1, 2, 3, 4, 5, 6]))

        s = str(a)
        self.assertIsInstance(s, str)
        self.assertEqual(s.count('\n'), 0)
        self.assertTrue(s.startswith('SpatialAcceleration'))

        r = np.random.rand(6, 10)
        a = SpatialAcceleration(r)
        self.assertIsInstance(a, SpatialAcceleration)
        self.assertIsInstance(a, SpatialVector)
        self.assertIsInstance(a, SpatialM6)
        self.assertEqual(len(a), 10)

        b = a[3]
        self.assertIsInstance(b, SpatialAcceleration)
        self.assertIsInstance(b, SpatialVector)
        self.assertIsInstance(b, SpatialM6)
        self.assertEqual(len(b), 1)
        self.assertTrue(all(b.A == r[:,3]))

        s = str(a)
        self.assertIsInstance(s, str)


    def test_force(self):

        a = SpatialForce([1, 2, 3, 4, 5, 6])
        self.assertIsInstance(a, SpatialForce)
        self.assertIsInstance(a, SpatialVector)
        self.assertIsInstance(a, SpatialF6)
        self.assertEqual(len(a), 1)
        self.assertTrue(all(a.A == np.r_[1, 2, 3, 4, 5, 6]))

        a = SpatialForce(np.r_[1, 2, 3, 4, 5, 6])
        self.assertIsInstance(a, SpatialForce)
        self.assertIsInstance(a, SpatialVector)
        self.assertIsInstance(a, SpatialF6)
        self.assertEqual(len(a), 1)
        self.assertTrue(all(a.A == np.r_[1, 2, 3, 4, 5, 6]))

        s = str(a)
        self.assertIsInstance(s, str)
        self.assertEqual(s.count('\n'), 0)
        self.assertTrue(s.startswith('SpatialForce'))

        r = np.random.rand(6, 10)
        a = SpatialForce(r)
        self.assertIsInstance(a, SpatialForce)
        self.assertIsInstance(a, SpatialVector)
        self.assertIsInstance(a, SpatialF6)
        self.assertEqual(len(a), 10)

        b = a[3]
        self.assertIsInstance(b, SpatialForce)
        self.assertIsInstance(b, SpatialVector)
        self.assertIsInstance(b, SpatialF6)
        self.assertEqual(len(b), 1)
        self.assertTrue(all(b.A == r[:, 3]))

        s = str(a)
        self.assertIsInstance(s, str)

    def test_momentum(self):

        a = SpatialMomentum([1, 2, 3, 4, 5, 6])
        self.assertIsInstance(a, SpatialMomentum)
        self.assertIsInstance(a, SpatialVector)
        self.assertIsInstance(a, SpatialF6)
        self.assertEqual(len(a), 1)
        self.assertTrue(all(a.A == np.r_[1, 2, 3, 4, 5, 6]))

        a = SpatialMomentum(np.r_[1, 2, 3, 4, 5, 6])
        self.assertIsInstance(a, SpatialMomentum)
        self.assertIsInstance(a, SpatialVector)
        self.assertIsInstance(a, SpatialF6)
        self.assertEqual(len(a), 1)
        self.assertTrue(all(a.A == np.r_[1, 2, 3, 4, 5, 6]))

        s = str(a)
        self.assertIsInstance(s, str)
        self.assertEqual(s.count('\n'), 0)
        self.assertTrue(s.startswith('SpatialMomentum'))

        r = np.random.rand(6, 10)
        a = SpatialMomentum(r)
        self.assertIsInstance(a, SpatialMomentum)
        self.assertIsInstance(a, SpatialVector)
        self.assertIsInstance(a, SpatialF6)
        self.assertEqual(len(a), 10)

        b = a[3]
        self.assertIsInstance(b, SpatialMomentum)
        self.assertIsInstance(b, SpatialVector)
        self.assertIsInstance(b, SpatialF6)
        self.assertEqual(len(b), 1)
        self.assertTrue(all(b.A == r[:, 3]))

        s = str(a)
        self.assertIsInstance(s, str)


    def test_arith(self):

        # just test SpatialVelocity since all types derive from same superclass

        r1 = np.r_[1, 2, 3, 4, 5, 6]
        r2 = np.r_[7, 8, 9, 10, 11, 12]
        a1 = SpatialVelocity(r1)
        a2 = SpatialVelocity(r2)

        self.assertTrue(all((a1 + a2).A == r1 + r2))
        self.assertTrue(all((a1 - a2).A == r1 - r2))
        self.assertTrue(all((-a1).A == -r1))

    def test_inertia(self):
        # constructor
        # addition
        pass

    def test_products(self):
        # v x v = a  *, v x F6 = a
        # a x I, I x a
        # v x I, I x v
        # twist x v, twist x a, twist x F
        pass


# ---------------------------------------------------------------------------------------#
if __name__ == '__main__':

    unittest.main()


import numpy as np
from spatialmath import base

# this is a collection of useful algorithms, not otherwise categorized

def numjac(f, x, dx=1e-8, SO=0, SE=0):
    r"""
    Numerically compute Jacobian of function

    :param f: the function, returns an m-vector
    :type f: callable
    :param x: function argument
    :type x: ndarray(n)
    :param dx: the numerical perturbation, defaults to 1e-8
    :type dx: float, optional
    :param SO: function returns SO(N) matrix, defaults to 0
    :type SO: int, optional
    :param SE: function returns SE(N) matrix, defaults to 0
    :type SE: int, optional

    :return: Jacobian matrix
    :rtype: ndarray(m,n)

    Computes a numerical approximation to the Jacobian for ``f(x)`` where
    :math:`f: \mathbb{R}^n \mapsto \mathbb{R}^m`.

    Uses first-order difference :math:`J[:,i] = (f(x + dx) - f(x)) / dx`.

    If ``SO`` is 2 or 3, then it is assumed that the function returns
    an SO(N) matrix and the derivative is converted to a column vector

    .. math:

        \vex \dmat{R} \mat{R}^T

    If ``SE`` is 2 or 3, then it is assumed that the function returns
    an SE(N) matrix and the derivative is converted to a colun vector.

    """
    x = np.array(x)
    Jcol = []
    J0 = f(x)
    I = np.eye(len(x))
    f0 = np.array(f(x))
    for i in range(len(x)):
        fi = np.array(f(x + I[:, i] * dx))
        Ji = (fi - f0) / dx

        if SE > 0:
            t = Ji[:SE, SE]
            r = base.vex(Ji[:SE, :SE] @ J0[:SE, :SE].T)
            Jcol.append(np.r_[t, r])
        elif SO > 0:
            R = Ji[:SO, :SO]
            r = base.vex(R @ J0[:SO, :SO].T)
            Jcol.append(r)
        else:
            Jcol.append(Ji)
        # print(Ji)

    return np.c_[Jcol].T

def numhess(J, x, dx=1e-8):
    r"""
    Numerically compute Hessian given Jacobian function

    :param J: the Jacobian function, returns an ndarray(m,n)
    :type J: callable
    :param x: function argument
    :type x: ndarray(n)
    :param dx: the numerical perturbation, defaults to 1e-8
    :type dx: float, optional
    :return: Hessian matrix
    :rtype: ndarray(m,n,n)

    Computes a numerical approximation to the Hessian for ``J(x)`` where 
    :math:`f: \mathbb{R}^n  \mapsto \mathbb{R}^{m \times n}`.

    The result is a 3D array where

    .. math::

        H_{i,j,k} = \frac{\partial J_{j,k}}{\partial x_i}

    Uses first-order difference :math:`H[:,:,i] = (J(x + dx) - J(x)) / dx`.
    """

    I = np.eye(len(x))
    Hcol = []
    J0 = J(x)
    for i in range(len(x)):

        Ji = J(x + I[:,i] * dx)
        Hi = (Ji - J0) / dx

        Hcol.append(Hi)
        
    return np.stack(Hcol, axis=0)

def array2str(X, valuesep=", ", rowsep=" | ", fmt="{:.3g}", 
    brackets=("[ ", " ]"), suppress_small=True):
    """
    Convert array to single line string

    :param X: 1D or 2D array to convert
    :type X: ndarray(N,M), array_like(N)
    :param valuesep: separator between numbers, defaults to ", "
    :type valuesep: str, optional
    :param rowsep: separator between rows, defaults to " | "
    :type rowsep: str, optional
    :param format: format string, defaults to "{:.3g}"
    :type precision: str, optional
    :param brackets: strings to be added to start and end of the string, 
        defaults to ("[ ", " ]").  Set to None to suppress brackets.
    :type brackets: list, tuple of str
    :param suppress_small: small values (:math:`|x| < 10^{-12}` are converted
        to zero, defaults to True
    :type suppress_small: bool, optional
    :return: compact string representation of array
    :rtype: str

    Converts a small array to a compact single line representation.
    """
    # convert to ndarray if not already
    if isinstance(X, (list, tuple)):
        X = base.getvector(X)
    
    def format_row(x):
        s = ""
        for j, e in enumerate(x):
            if abs(e) < 1e-12:
                e = 0
            if j > 0:
                s += valuesep
            s += fmt.format(e)
        return s
    
    if X.ndim == 1:
        # 1D case
        s = format_row(X)
    else:
        # 2D case
        s = ""
        for i, row in enumerate(X):
            if i > 0:
                s += rowsep
            s += format_row(row)

    if brackets is not None and len(brackets) == 2:
        s = brackets[0] + s + brackets[1]
    return s

def bresenham(p0, p1, array=None):
    """
    Line drawing in a grid

    :param p0: initial point
    :type p0: array_like(2) of int
    :param p1: end point
    :type p1: array_like(2) of int
    :return: arrays of x and y coordinates for points along the line
    :rtype: ndarray(N), ndarray(N) of int

    Return x and y coordinate vectors for points in a grid that lie on
    a line from ``p0`` to ``p1`` inclusive.

    The end points, and all points along the line are integers.

    .. note:: The API is similar to the Bresenham algorithm but this
        implementation uses NumPy vectorised arithmetic which makes it 
        faster than the Bresenham algorithm in Python.
    """
    x0, y0 = p0
    x1, y1 = p1

    if array is not None:
        _ = array[y0, x0] + array[y1, x1]

    dx = x1 - x0
    dy = y1 - y0

    if abs(dx) >= abs(dy):
        # shallow line -45° <= θ <= 45°
        # y = mx + c
        if dx == 0:
            # case p0 == p1
            x = np.r_[x0]
            y = np.r_[y0]
        else:
            m = dy / dx
            c = y0 - m * x0
            if dx > 0:
                # line to the right
                x = np.arange(x0, x1 + 1)
            elif dx < 0:
                # line to the left
                x = np.arange(x0, x1 - 1, -1)
            y = np.round(x * m + c)

    else:
        # steep line  θ < -45°,  θ > 45°
        # x = my + c
        m = dx / dy
        c = x0 - m * y0
        if dy > 0:
            # line to the right
            y = np.arange(y0, y1 + 1)
        elif dy < 0:
            # line to the left
            y = np.arange(y0, y1 - 1, -1)
        x = np.round(y * m + c)

    return x.astype(int), y.astype(int)

def mpq_point(data, p, q):
    r"""
    Moments of polygon

    :param p: moment order x
    :type p: int
    :param q: moment order y
    :type q: int

    Returns the pq'th moment of the polygon

    .. math::
    
        M(p, q) = \sum_{i=0}^{n-1} x_i^p y_i^q

    Example:

    .. runblock:: pycon

        >>> from spatialmath import Polygon2
        >>> p = Polygon2([[1, 3, 2], [2, 2, 4]])
        >>> p.moment(0, 0)  # area
        >>> p.moment(3, 0)

    Note is negative for clockwise perimeter.
    """
    x = data[0, :]
    y = data[1, :]

    return np.sum(x**p * y**q)

def gauss1d(mu, var, x):
    """
    Gaussian function in 1D

    :param mu: mean
    :type mu: float
    :param var: variance
    :type var: float
    :param x: x-coordinate values
    :type x: array_like(n)
    :return: Gaussian :math:`G(x)`
    :rtype: ndarray(n)

    :seealso: :func:`gauss2d`
    """
    sigma = np.sqrt(var)
    x = base.getvector(x)

    return 1.0 / np.sqrt(sigma**2 * 2 * np.pi) * np.exp(-(x-mu)**2/2/sigma**2)

def gauss2d(mu, P, X, Y):
    """
    Gaussian function in 2D

    :param mu: mean
    :type mu: array_like(2)
    :param P: covariance matrix
    :type P: ndarray(2,2)
    :param X: array of x-coordinates
    :type X:  ndarray(n,m)
    :param Y: array of y-coordinates
    :type Y: ndarray(n,m)
    :return: Gaussian :math:`g(x,y)`
    :rtype: ndarray(n,m)

    Computed :math:`g_{i,j} = G(x_{i,j}, y_{i,j})`

    :seealso: :func:`gauss1d`
    """

    x = X.ravel() - mu[0]
    y = Y.ravel() - mu[1]

    Pi = np.linalg.inv(P);
    g = 1/(2*np.pi*np.sqrt(np.linalg.det(P))) * np.exp(
         -0.5*(x**2 * Pi[0, 0] + y**2 * Pi[1, 1] + 2 * x * y * Pi[0, 1]));
    return g.reshape(X.shape)

if __name__ == "__main__":

    r = np.linspace(-4, 4, 6)
    x, y = np.meshgrid(r, r)
    print(gauss2d([0, 0], np.diag([1,2]), x, y))
    # print(bresenham([2,2], [2,4]))
    # print(bresenham([2,2], [2,-4]))
    # print(bresenham([2,2], [4,2]))
    # print(bresenham([2,2], [-4,2]))
    # print(bresenham([2,2], [2,2]))
    # print(bresenham([2,2], [3,6])) # steep
    # print(bresenham([2,2], [6,3])) # shallow
    # print(bresenham([2,2], [3,6])) # steep
    # print(bresenham([2,2], [6,3])) # shallow


from spatialmath.pose2d import SO2, SE2
from spatialmath.pose3d import SO3, SE3
from spatialmath.baseposematrix import BasePoseMatrix
from spatialmath.geom2d import Line2, Polygon2
from spatialmath.geom3d import Line3, Plane3
from spatialmath.twist import Twist3, Twist2
from spatialmath.spatialvector import SpatialVelocity, SpatialAcceleration, \
     SpatialForce, SpatialMomentum, SpatialInertia
from spatialmath.quaternion import Quaternion, UnitQuaternion
from spatialmath.DualQuaternion import DualQuaternion, UnitDualQuaternion
#from spatialmath.Plucker import *
from spatialmath import base as smb


__all__ = [
    # pose
    "SO2",
    "SE2",
    "SO3",
    "SE3",
    "BasePoseMatrix",
    "Quaternion",
    "UnitQuaternion",
    "DualQuaternion",
    "UnitDualQuaternion",
    "Twist3",
    "Twist2",
    "SpatialVelocity",
    "SpatialAcceleration",
    "SpatialForce",
    "SpatialMomentum",
    "SpatialInertia",
    "Line3",
    "Plane3",
    "Line2",
    "Polygon2",
    "smb",
]





import cProfile
import spatialmath as sm

def myfunc():
    X1 = sm.SE3.Rand()
    X2 = sm.SE3.Rand()
    Y = []      
    for i in range(1000000):
        Y.append( X1 * X2 )


cProfile.run('myfunc')


#!/usr/bin/env python3
# -*- coding", t)
"""
Created on Fri Apr 10 14:22:36 2020

@author", t)
"""


import timeit
from ansitable import ANSITable, Column
import numpy as np
from spatialmath import base

N = 100000

table = ANSITable(
    Column("Operation", headalign="^"),
    Column("Time (μs)", headalign="^", fmt="{:.2f}"),
    border="thick",
)


def result(op, t):
    global table

    table.row(op, t / N * 1e6)


# ------------------------------------------------------------------------- #
setup = """
from spatialmath import base
from spatialmath.base import r2q
import numpy as np
s = np.r_[1.0,2,3,4,5,6]
s3 = np.r_[1.0,2,3]
a = np.r_[1.0, 2.0, 3.0]
b = np.r_[-5.0, 4.0, 3.0]
R = base.eul2r(0.1, 0.2, 0.3)
"""
table.rule()
from spatialmath.base import r2q #, r2q_old


t = timeit.timeit(stmt="q = r2q(R)", setup=setup, number=N)
result("r2q()", t)

# t = timeit.timeit(stmt='q = r2q_old(R)', setup=setup, number=N)
# result("r2q_old()", t)

t = timeit.timeit(stmt="c = np.cross(a,b)", setup=setup, number=N)
result("np.cross()", t)

t = timeit.timeit(stmt="c = base.cross(a,b)", setup=setup, number=N)
result("cross()", t)

# t = timeit.timeit(stmt="a = np.inner(s,s).sum()", setup=setup, number=N)
# result("inner()", t)

# t = timeit.timeit(stmt="a = np.linalg.norm(s) ** 2", setup=setup, number=N)
# result("np.norm**2", t)

# t = timeit.timeit(stmt="a = base.normsq(s)", setup=setup, number=N)
# result("base.normsq", t)

# t = timeit.timeit(stmt="a = (s ** 2).sum()", setup=setup, number=N)
# result("s**2.sum()", t)

# t = timeit.timeit(stmt="a = np.sum(s ** 2)", setup=setup, number=N)
# result("np.sum(s ** 2)", t)

# t = timeit.timeit(stmt="a = np.linalg.norm(s)", setup=setup, number=N)
# result("np.norm(R6)", t)
# t = timeit.timeit(stmt="a = base.norm(s)", setup=setup, number=N)
# result("base.norm(R6)", t)

# t = timeit.timeit(stmt="a = np.linalg.norm(s3)", setup=setup, number=N)
# result("np.norm(R3)", t)

# t = timeit.timeit(stmt="a = base.norm(s3)", setup=setup, number=N)
# result("base.norm(R3)", t)

# t = timeit.timeit(stmt="a = np.isclose(s, s)", setup=setup, number=N)
# result("isclose(s,s)", t)

# t = timeit.timeit(stmt="a = base.iszerovec(s-s)", setup=setup, number=N)
# result("base.iszerovec(s-s)", t)


t = timeit.timeit(stmt="RR=np.eye(4); RR[:3,:3]=R", setup=setup, number=N)
result("RR=np.eye(4); RR[:3,:3]=R", t)

t = timeit.timeit(stmt="RR=np.pad(R, ((0,1),(0,1)), 'constant')", setup=setup, number=N)
result("RR=np.pad(R, ((0,1),(0,1)), 'constant')", t)



table.print()



#!/usr/bin/env python3
# -*- coding", t)
"""
Created on Fri Apr 10 14:22:36 2020

@author", t)
"""


import timeit
from ansitable import ANSITable, Column

N = 100000

table = ANSITable(
    Column("Operation", headalign="^"),
    Column("Time (μs)", headalign="^", fmt="{:.2f}"),
    border="thick")

def result(op, t):
    global table

    table.row(op, t/N*1e6)
# ------------------------------------------------------------------------- #

# transforms_setup = '''
# from spatialmath import SE3
# from spatialmath import base

# import numpy as np
# from collections import namedtuple
# Rt = namedtuple('Rt', 'R t')
# X1 = SE3.Rand()
# X2 = SE3.Rand()
# T1 = X1.A
# T2 = X2.A
# R1 = base.t2r(T1)
# R2 = base.t2r(T2)
# t1 = base.transl(T1)
# t2 = base.transl(T2)
# Rt1 = Rt(R1, t1)
# Rt2 = Rt(R2, t2)
# v = np.r_[1,2,3]
# v2 = np.r_[1,2,3, 1]
# '''
# t = timeit.timeit(stmt='base.getvector(0.2)', setup=transforms_setup, number=N)
# result("getvector(x)", t)

# t = timeit.timeit(stmt='base.rotx(0.2, unit="rad")', setup=transforms_setup, number=N)
# result("base.rotx", t)

# t = timeit.timeit(stmt='base.trotx(0.2, unit="rad")', setup=transforms_setup, number=N)
# result("base.trotx", t)

# t = timeit.timeit(stmt='base.t2r(T1)', setup=transforms_setup, number=N)
# result("base.t2r", t)

# t = timeit.timeit(stmt='base.r2t(R1)', setup=transforms_setup, number=N)
# result("base.r2t", t)

# t = timeit.timeit(stmt='T1 @ T2', setup=transforms_setup, number=N)
# result("4x4 @", t)

# t = timeit.timeit(stmt='T1[:3,:3] @ T2[:3,:3] + T1[:3,:3] @ T2[:3,3]', setup=transforms_setup, number=N)
# result("R1*R2, R1*t", t)

# t = timeit.timeit(stmt='(Rt1.R @ Rt2.R, Rt1.R @ Rt2.t)', setup=transforms_setup, number=N)
# result("T1 * T2 (R, t)", t)

# t = timeit.timeit(stmt='base.trinv(T1)', setup=transforms_setup, number=N)
# result("base.trinv", t)

# t = timeit.timeit(stmt='(Rt1.R.T, -Rt1.R.T @ Rt1.t)', setup=transforms_setup, number=N)
# result("base.trinv (R,t)", t)

# t = timeit.timeit(stmt='np.linalg.inv(T1)', setup=transforms_setup, number=N)
# result("np.linalg.inv", t)

# t = timeit.timeit(stmt='T1 @ v2', setup=transforms_setup, number=N)
# result("(4,4) * (4,)", t)

# # ------------------------------------------------------------------------- #
# table.rule()

# t = timeit.timeit(stmt='SE3()', setup=transforms_setup, number=N)
# result("SE3()", t)

# t = timeit.timeit(stmt='SE3.Rx(0.2)', setup=transforms_setup, number=N)
# result("SE3.Rx()", t)

# t = timeit.timeit(stmt='T1[:3,:3]', setup=transforms_setup, number=N)
# result("T1[:3,:3]", t)

# t = timeit.timeit(stmt='X1.A', setup=transforms_setup, number=N)
# result("SE3.A", t)

# t = timeit.timeit(stmt='SE3(T1)', setup=transforms_setup, number=N)
# result("SE3(T1)", t)

# t = timeit.timeit(stmt='SE3(T1, check=False)', setup=transforms_setup, number=N)
# result("SE3(T1 check=False)", t)

# t = timeit.timeit(stmt='SE3([T1], check=False)', setup=transforms_setup, number=N)
# result("SE3([T1])", t)

# t = timeit.timeit(stmt='X1 * X2', setup=transforms_setup, number=N)
# result("SE3 * SE3", t)

# t = timeit.timeit(stmt='X1.inv()', setup=transforms_setup, number=N)
# result("SE3.inv", t)

# t = timeit.timeit(stmt='X1 * v', setup=transforms_setup, number=N)
# result("SE3 * v", t)

# t = timeit.timeit(stmt='a = X1.log()', setup=transforms_setup, number=N)
# result("SE3.log()", t)

# # ------------------------------------------------------------------------- #
# quat_setup = '''
# from spatialmath import base
# from spatialmath import UnitQuaternion
# import numpy as np
# q1 = base.rand()
# q2 = base.rand()
# v = np.r_[1,2,3]
# Q1 = UnitQuaternion.Rx(0.2)
# Q2 = UnitQuaternion.Ry(0.3)
# '''
# table.rule()

# t = timeit.timeit(stmt='a = UnitQuaternion()', setup=quat_setup, number=N)
# result("UnitQuaternion() ", t)

# t = timeit.timeit(stmt='a = UnitQuaternion.Rx(0.2)', setup=quat_setup, number=N)
# result("UnitQuaternion.Rx ", t)

# t = timeit.timeit(stmt='a = Q1 * Q2', setup=quat_setup, number=N)
# result("UnitQuaternion * UnitQuaternion", t)

# t = timeit.timeit(stmt='a = Q1 * v', setup=quat_setup, number=N)
# result("UnitQuaternion * v", t)

# t = timeit.timeit(stmt='a = base.qqmul(q1,q2)', setup=quat_setup, number=N)
# result("base.qqmul", t)

# t = timeit.timeit(stmt='a = base.qvmul(q1,v)', setup=quat_setup, number=N)
# result("base.qvmul", t)



# # ------------------------------------------------------------------------- #
# twist_setup = '''
# from spatialmath import SE3, Twist3
# from spatialmath import base
# import numpy as np
# from math import cos
# S1 = SE3.Rand().Twist3()
# S2 = SE3.Rand().Twist3()
# X1 = SE3.Rand()
# T1 = X1.A
# A1 = X1.Ad()
# se3 = S1.se3()
# s = np.r_[1,2,3,4,5,6]
# v = np.r_[1,2,3]
# '''
# table.rule()
# t = timeit.timeit(stmt='a = Twist3()', setup=twist_setup, number=N)
# result("Twist3()", t)

# t = timeit.timeit(stmt='a = X1.Twist3()', setup=twist_setup, number=N)
# result("SE3.Twist3()", t)

# t = timeit.timeit(stmt='a = S1 * S2', setup=twist_setup, number=N)
# result("Twist3 * Twist3", t)

# t = timeit.timeit(stmt='a = S1.inv()', setup=twist_setup, number=N)
# result("Twist3.inv()", t)

# t = timeit.timeit(stmt='a = S1.Ad()', setup=twist_setup, number=N)
# result("Twist3.Ad()", t)

# t = timeit.timeit(stmt='a = S1.exp(1)', setup=twist_setup, number=N)
# result("Twist3.Exp()", t)

# t = timeit.timeit(stmt='a = base.skewa(v)', setup=twist_setup, number=N)
# result("skew", t)

# t = timeit.timeit(stmt='a = base.skewa(s)', setup=twist_setup, number=N)
# result("skewa", t)

# t = timeit.timeit(stmt='a = base.vexa(se3)', setup=twist_setup, number=N)
# result("vexa", t)

# t = timeit.timeit(stmt='a = base.trlog(T1)', setup=twist_setup, number=N)
# result("trlog", t)

# t = timeit.timeit(stmt='a = base.trlog(T1, twist=True)', setup=twist_setup, number=N)
# result("trlog as twist", t)

# t = timeit.timeit(stmt='a = base.trexp(se3)', setup=twist_setup, number=N)
# result("trexp", t)

# t = timeit.timeit(stmt='a = A1 @ s', setup=twist_setup, number=N)
# result("(6,6) * (6,)", t)

# t = timeit.timeit(stmt='a = base.rodrigues(v)', setup=twist_setup, number=N)
# result("rodrigues", t)

# t = timeit.timeit(stmt='a = cos(0.3)', setup=twist_setup, number=N)
# result("math.cos", t)

# t = timeit.timeit(stmt='a = np.cos(0.3)', setup=twist_setup, number=N)
# result("np.cos", t)

# ------------------------------------------------------------------------- #
misc_setup = '''
from spatialmath import base
import numpy as np
s = np.r_[1.0,2,3,4,5,6]
s3 = np.r_[1.0,2,3]
a = np.r_[1.0, 2.0, 3.0]
b = np.r_[-5.0, 4.0, 3.0]

A = np.random.randn(6,6)
As = (A + A.T) / 2
bb = np.random.randn(6)
'''
table.rule()

t = timeit.timeit(stmt='c = np.linalg.inv(As)', setup=misc_setup, number=N)
result("np.inv(As)", t)

t = timeit.timeit(stmt='c = np.linalg.pinv(As)', setup=misc_setup, number=N)
result("np.pinv(As)", t)

t = timeit.timeit(stmt='c = np.linalg.solve(As, bb)', setup=misc_setup, number=N)
result("np.solve(As, b)", t)

t = timeit.timeit(stmt='c = np.cross(a,b)', setup=misc_setup, number=N)
result("np.cross()", t)

t = timeit.timeit(stmt='c = base.cross(a,b)', setup=misc_setup, number=N)
result("cross()", t)

t = timeit.timeit(stmt='a = np.inner(s,s).sum()', setup=misc_setup, number=N)
result("inner()", t)

t = timeit.timeit(stmt='a = np.linalg.norm(s) ** 2', setup=misc_setup, number=N)
result("np.norm**2", t)

t = timeit.timeit(stmt='a = base.normsq(s)', setup=misc_setup, number=N)
result("base.normsq", t)

t = timeit.timeit(stmt='a = (s ** 2).sum()', setup=misc_setup, number=N)
result("s**2.sum()", t)

t = timeit.timeit(stmt='a = np.sum(s ** 2)', setup=misc_setup, number=N)
result("np.sum(s ** 2)", t)

t = timeit.timeit(stmt='a = np.linalg.norm(s)', setup=misc_setup, number=N)
result("np.norm(R6)", t)
t = timeit.timeit(stmt='a = base.norm(s)', setup=misc_setup, number=N)
result("base.norm(R6)", t)

t = timeit.timeit(stmt='a = np.linalg.norm(s3)', setup=misc_setup, number=N)
result("np.norm(R3)", t)
t = timeit.timeit(stmt='a = base.norm(s3)', setup=misc_setup, number=N)
result("base.norm(R3)", t)


table.print()


# Part of Spatial Math Toolbox for Python
# Copyright (c) 2000 Peter Corke
# MIT Licence, see details in top-level file: LICENCE

# matplotlib inline

# line.set_data()
# text.set_position()
# quiver.set_offsets(), quiver.set_UVC()
# FancyArrow.set_xy()
import os.path
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import animation
from spatialmath import base
from collections.abc import Iterable, Iterator

# global variable holds reference to FuncAnimation object, this is essential
# for animatiion to work
_ani = None


class Animate:
    """
    Animate objects for matplotlib 3d

    An instance of this class behaves like an Axes3D and supports proxies for

    - ``plot``
    - ``quiver``
    - ``text``

    which renders them and also places corresponding objects into a display
    list. These objects are ``Line``, ``Quiver`` and ``Text``.  Only these
    primitives will be animated.

    The objects are all drawn relative to the origin, and will be transformed
    according to the transform that is being animated.

    Example::

        anim = animate.Animate(dims=[0,2]) # set up the 3D axes
        anim.trplot(T, frame='A', color='green')  # draw the frame
        anim.run(loop=True)  # animate it
    """

    def __init__(
        self, axes=None, dims=None, projection="ortho", labels=("X", "Y", "Z"), **kwargs
    ):
        """
        Construct an Animate object

        :param axes: the axes to plot into, defaults to current axes
        :type axes: Axes3D reference
        :param dims: dimension of plot volume as [xmin, xmax, ymin, ymax,
            zmin, zmax]. If dims is [min, max] those limits are applied
            to the x-, y- and z-axes.
        :type dims: array_like(6) or array_like(2)
        :param projection: 3D projection: ortho [default] or persp
        :type projection: str
        :param labels: labels for the axes, defaults to X, Y and Z
        :type labels: 3-tuple of strings

        Will setup to plot into an existing or a new Axes3D instance.

        """
        self.trajectory = None
        self.displaylist = []

        if axes is None:
            # no axes specified

            fig = plt.gcf()
            # check any current axes
            for a in fig.axes:
                if a.name != "3d":
                    # if they are not 3D axes, remove them, otherwise will
                    # get plot errors
                    a.remove()
            if len(fig.axes) == 0:
                # no axes in the figure, create a 3D axes
                axes = fig.add_subplot(111, projection="3d", proj_type=projection)
                axes.set_xlabel(labels[0])
                axes.set_ylabel(labels[1])
                axes.set_zlabel(labels[2])
                axes.autoscale(enable=True, axis="both")
            else:
                # reuse an existing axis
                axes = plt.gca()

        if dims is not None:
            if len(dims) == 2:
                dims = dims * 3
            axes.set_xlim(dims[0:2])
            axes.set_ylim(dims[2:4])
            axes.set_zlim(dims[4:6])
            # ax.set_aspect('equal')

        self.ax = axes

        # TODO set flag for 2d or 3d axes, flag errors on the methods called later

    def trplot(self, end, start=None, **kwargs):
        """
        Define the transform to animate

        :param end: the final pose SE(3) or SO(3) to display as a coordinate frame
        :type end: ndarray(4,4) or ndarray(3,3)
        :param start: the initial pose SE(3) or SO(3) to display as a coordinate frame, defaults to null
        :type start: ndarray(4,4) or ndarray(3,3)
        :param start: an

        Is polymorphic with ``base.trplot`` and accepts the same parameters.
        This sets up the animation but doesn't execute it.

        :seealso: :func:`run`

        """
        self.trajectory = None
        if not isinstance(end, (np.ndarray, np.generic)) and isinstance(end, Iterable):
            try:
                if len(end) == 1:
                    end = end[0]
                elif len(end) >= 2:
                    self.trajectory = end
            except TypeError:
                # a generator has no len()
                self.trajectory = end

        # stash the final value
        if base.isrot(end):
            self.end = base.r2t(end)
        else:
            self.end = end

        if start is None:
            self.start = np.identity(4)
        else:
            if base.isrot(start):
                self.start = base.r2t(start)
            else:
                self.start = start

        # draw axes at the origin
        base.trplot(self.start, ax=self, **kwargs)

    def set_proj_type(self, proj_type):
        self.ax.set_proj_type(proj_type)

    def run(
        self,
        movie=None,
        axes=None,
        repeat=False,
        interval=50,
        nframes=100,
        wait=False,
        **kwargs
    ):
        """
        Run the animation

        :param axes: the axes to plot into, defaults to current axes
        :type axes: Axes3D reference
        :param repeat: animate in endless loop [default False]
        :type repeat: bool
        :param nframes: number of steps in the animation [default 100]
        :type nframes: int
        :param interval: number of milliseconds between frames [default 50]
        :type interval: int
        :param movie: name of file to write MP4 movie into
        :type movie: str
        :param wait: wait until animation is complete, default False
        :type wait: bool

        Animates a 3D coordinate frame moving from the world frame to a frame
        represented by the SO(3) or SE(3) matrix to the current axes.

        .. note::

            - the ``movie`` option requires the ffmpeg package to be installed:
              ``conda install -c conda-forge ffmpeg``
            - invokes the draw() method of every object in the display list
        """

        def update(frame, animation):

            # frame is the result of calling next() on a iterator or generator
            # seemingly the animation framework isn't checking StopException
            # so there is no way to know when this is no longer called.
            # we implement a rather hacky heartbeat style timeout

            if isinstance(frame, float):
                # passed a single transform, interpolate it
                T = base.trinterp(start=self.start, end=self.end, s=frame)
            else:
                # assume it is an SO(3) or SE(3)
                T = frame

            # ensure result is SE(3)
            if T.shape == (3, 3):
                T = base.r2t(T)

            # update the scene
            animation._draw(T)

            self.count += 1  # say we're still running

            return animation.artists()

        global _ani

        # blit leaves a trail and first frame
        if movie is not None:
            repeat = False

        self.count = 1
        if self.trajectory is not None:
            if not isinstance(self.trajectory, Iterator):
                # make it iterable, eg. if a list or tuple
                self.trajectory = iter(self.trajectory)
            frames = self.trajectory
        else:
            frames = iter(np.linspace(0, 1, nframes))

        _ani = animation.FuncAnimation(
            fig=plt.gcf(),
            func=update,
            frames=frames,
            fargs=(self,),
            blit=False,
            interval=interval,
            repeat=repeat,
        )

        if movie is not None:
            # Set up formatting for the movie files
            if os.path.exists(movie):
                print("overwriting movie", movie)
            else:
                print("creating movie", movie)
            FFwriter = animation.FFMpegWriter(
                fps=1000 / interval, extra_args=["-vcodec", "libx264"]
            )
            _ani.save(movie, writer=FFwriter)

        if wait:
            # wait for the animation to finish.  Dig into the timer for this
            # animation and wait for its callback to be deregistered.
            while True:
                plt.pause(0.25)
                if _ani.event_source is None or len(_ani.event_source.callbacks) == 0:
                    break

    def __repr__(self):
        """
        Human readable version of the display list

        :param self: the animation
        :type self: Animate
        :returns: readable version of the display list
        :rtype: str
        """
        return "Animate(" + ", ".join([x.type for x in self.displaylist]) + ")"

    def __str__(self):
        return f"Animate(len={len(self.displaylist)}"

    def artists(self):
        """
        List of artists that need to be updated

        :param self: the animation
        :type self: Animate
        :returns: list of artists
        :rtype: list
        """
        return [x.h for x in self.displaylist]

    def _draw(self, T):
        for x in self.displaylist:
            x.draw(T)

    # ------------------- plot()

    class _Line:
        def __init__(self, anim, h, xs, ys, zs):
            # form 4xN matrix, columns are first/last point in homogeneous form
            p = np.vstack([xs, ys, zs])
            self.p = np.vstack([p, np.ones((p.shape[1],))])
            self.h = h
            self.type = "line"
            self.anim = anim

        def draw(self, T):
            p = T @ self.p
            self.h.set_data(p[0, :], p[1, :])
            self.h.set_3d_properties(p[2, :])

    def plot(self, x, y, z, *args, **kwargs):
        """
        Plot a polyline

        :param x: list of x-coordinates
        :type x: array_like
        :param y: list of y-coordinates
        :type y: array_like
        :param z: list of z-coordinates
        :type z: array_like

        Other arguments as accepted by the matplotlib method.

        All arrays must have the same length.

        :seealso: :func:`matplotlib.pyplot.plot`
        """

        (h,) = self.ax.plot(x, y, z, *args, **kwargs)
        self.displaylist.append(Animate._Line(self, h, x, y, z))
        return h

    # ------------------- quiver()

    class _Quiver:
        def __init__(self, anim, h):
            self.type = "quiver"
            self.anim = anim
            # for matplotlib 3.1.x
            # ._segments3d is 3x2x3
            #   first index: line segment in the collection
            #   second index: 0 = start, 1 = end
            #   third index: x, y, z components
            # https://stackoverflow.com/questions/48911643/set-uvc-equivilent-for-a-3d-quiver-plot-in-matplotlib
            #
            # for matplotlib 3.3.x
            # ._segments3d is a 3-element list, each element is 2x3

            # turn to homogeneous form, with columns per point, alternating start, end

            if isinstance(h._segments3d, np.ndarray):
                self.p = np.vstack(
                    [h._segments3d.reshape(6, 3).T, np.ones((1, 6))]
                )  # result is 4x6
            else:
                self.p = np.vstack(
                    [np.hstack([x.T for x in h._segments3d]), np.ones((1, 6))]
                )
            self.h = h
            self.type = "arrow"
            self.anim = anim

        def draw(self, T):
            p = T @ self.p

            # reshape it
            p = p[0:3, :].T.reshape(3, 2, 3)
            self.h.set_segments(p)

    def quiver(self, x, y, z, u, v, w, *args, **kwargs):
        """
        Plot a quiver

        :param x: list of base x-coordinates
        :type x: array_like
        :param y: list of base y-coordinates
        :type y: array_like
        :param z: list of base z-coordinates
        :type z: array_like
        :param u: list of vector x-coordinates
        :type u: array_like
        :param v: list of vector y-coordinates
        :type v: array_like
        :param w: list of vector z-coordinates
        :type w: array_like

        Draws a series of arrows, the bases defined by corresponding elements
        of (x,y,z) and the vector has components defined by corresponding
        elements of (u,v,w).

        Other arguments as accepted by the matplotlib method.

        :seealso: :func:`matplotlib.pyplot.quiver`
        """
        h = self.ax.quiver(x, y, z, u, v, w, *args, **kwargs)
        self.displaylist.append(Animate._Quiver(self, h))

    # ------------------- text()

    class _Text:
        def __init__(self, anim, h, x, y, z):
            self.type = "text"
            self.h = h
            self.p = np.r_[x, y, z, 1]
            self.anim = anim

        def draw(self, T):
            p = T @ self.p
            # x2, y2, _ = proj3d.proj_transform(
            #   p[0], p[1], p[2], self.anim.ax.get_proj())
            # self.h.set_position((x2, y2))
            self.h.set_position((p[0], p[1]))
            self.h.set_3d_properties(z=p[2], zdir="x")

    def text(self, x, y, z, *args, **kwargs):
        """
        Plot text

        :param x: x-coordinate
        :type x: float
        :param y: float
        :type y: float
        :param z: z-coordinate
        :type z: float
        :param kwargs: Other arguments as accepted by the matplotlib method.

        ``.text(x, y, z, s)`` display the string ``s`` at coordinate
        (``x``, ``y``, ``z``).

        :seealso: :func:`~matplotlib.pyplot.text`
        """
        h = self.ax.text3D(x, y, z, *args, **kwargs)
        self.displaylist.append(Animate._Text(self, h, x, y, z))

    # ------------------- scatter()

    def scatter(self, xs, ys, zs, s=0, **kwargs):
        h = self.plot(xs, ys, zs, '.', markersize=0, **kwargs)
        self.displaylist.append(Animate._Line(self, h, xs, ys, zs))

    # ------------------- wrappers for Axes primitives

    def set_xlim(self, *args, **kwargs):
        self.ax.set_xlim(*args, **kwargs)

    def set_ylim(self, *args, **kwargs):
        self.ax.set_ylim(*args, **kwargs)

    def set_zlim(self, *args, **kwargs):
        self.ax.set_zlim(*args, **kwargs)

    def set_xlabel(self, *args, **kwargs):
        self.ax.set_xlabel(*args, **kwargs)

    def set_ylabel(self, *args, **kwargs):
        self.ax.set_ylabel(*args, **kwargs)

    def set_zlabel(self, *args, **kwargs):
        self.ax.set_zlabel(*args, **kwargs)


class Animate2:
    """
    Animate objects for matplotlib 2d

    An instance of this class behaves like an Axes3D and supports proxies for

    - ``plot``
    - ``quiver``
    - ``text``

    which renders them and also places corresponding objects into a display
    list. These objects are ``Line``, ``Quiver`` and ``Text``.  Only these
    primitives will be animated.

    The objects are all drawn relative to the origin, and will be transformed
    according to the transform that is being animated.

    Example::

        anim = animate.Animate(dims=[0,2]) # set up the 3D axes
        anim.trplot(T, frame='A', color='green')  # draw the frame
        anim.run(loop=True)  # animate it
    """

    def __init__(self, axes=None, dims=None, labels=("X", "Y"), **kwargs):
        """
        Construct an Animate object

        :param axes: the axes to plot into, defaults to current axes
        :type axes: Axes3D reference
        :param dims: dimension of plot volume as [xmin, xmax, ymin, ymax]. If
            dims is [min, max] those limits are applied to the x- and y-axes.
        :type dims: array_like(4) or array_like(2)
        :param projection: 3D projection: ortho [default] or persp
        :type projection: str
        :param labels: labels for the axes, defaults to X, Y and Z
        :type labels: 3-tuple of strings

        Will setup to plot into an existing or a new Axes3D instance.

        """
        self.trajectory = None
        self.displaylist = []

        if axes is None:
            # create an axes
            fig = plt.gcf()
            if fig.axes is None:
                # no axes in the figure, create a 3D axes
                axes = fig.add_subplot(111)
                axes.set_xlabel(labels[0])
                axes.set_ylabel(labels[1])
                axes.autoscale(enable=True, axis="both")
            else:
                # reuse an existing axis
                axes = plt.gca()

        if dims is not None:
            if len(dims) == 2:
                dims = dims * 2
            axes.set_xlim(dims[0:2])
            axes.set_ylim(dims[2:4])
            # ax.set_aspect('equal')
        else:
            axes.autoscale(enable=True, axis='both')

        self.ax = axes

        # set flag for 2d or 3d axes, flag errors on the methods called later

    def trplot2(self, end, start=None, **kwargs):
        """
        Define the transform to animate

        :param end: the final pose SE(2) or SO(2) to display as a coordinate frame
        :type end: ndarray(3,3) or ndarray(2,2)
        :param start: the initial pose SE(2) or SO(2) to display as a coordinate frame, defaults to null
        :type start: ndarray(3,3) or ndarray(2,2)

        Is polymorphic with ``base.trplot`` and accepts the same parameters.
        This sets up the animation but doesn't execute it.

        :seealso: :func:`run`

        """
        if not isinstance(end, (np.ndarray, np.generic)) and isinstance(end, Iterable):
            if len(end) == 1:
                end = end[0]
            elif len(end) >= 2:
                self.trajectory = end

        # stash the final value
        if base.isrot2(end):
            self.end = base.r2t(end)
        else:
            self.end = end

        if start is None:
            self.start = np.identity(3)
        else:
            if base.isrot2(start):
                self.start = base.r2t(start)
            else:
                self.start = start

        # draw axes at the origin
        base.trplot2(self.start, ax=self, block=False, **kwargs)

    def run(
        self, movie=None, axes=None, repeat=False, interval=50, nframes=100, **kwargs
    ):
        """
        Run the animation

        :param axes: the axes to plot into, defaults to current axes
        :type axes: Axes3D reference
        :param nframes: number of steps in the animation [defaault 100]
        :type nframes: int
        :param repeat: animate in endless loop [default False]
        :type repeat: bool
        :param interval: number of milliseconds between frames [default 50]
        :type interval: int
        :param movie: name of file to write MP4 movie into
        :type movie: str

        Animates a 3D coordinate frame moving from the world frame to a frame
        represented by the SO(3) or SE(3) matrix to the current axes.

        .. note::

            - the ``movie`` option requires the ffmpeg package to be installed:
              ``conda install -c conda-forge ffmpeg``
            - invokes the draw() method of every object in the display list
        """

        def update(frame, a):
            # if contains trajectory:
            if self.trajectory is not None:
                T = self.trajectory[frame]
            else:
                T = base.trinterp2(start=self.start, end=self.end, s=frame / nframes)
            a._draw(T)
            if frame == nframes - 1:
                a.done = True
            return a.artists()

        # blit leaves a trail and first frame
        if movie is not None:
            repeat = False

        self.done = False
        if self.trajectory is not None:
            nframes = len(self.trajectory)
        ani = animation.FuncAnimation(
            fig=plt.gcf(),
            func=update,
            frames=range(0, nframes),
            fargs=(self,),
            blit=False,
            interval=interval,
            repeat=repeat,
        )

        if movie is None:
            while repeat or not self.done:
                plt.pause(1)
        else:
            # Set up formatting for the movie files
            if os.path.exists(movie):
                print("overwriting movie", movie)
            else:
                print("creating movie", movie)
            FFwriter = animation.FFMpegWriter(fps=10, extra_args=["-vcodec", "libx264"])
            ani.save(movie, writer=FFwriter)

    def __repr__(self):
        """
        Human readable version of the display list

        :param self: the animation
        :type self: Animate
        :returns: readable version of the display list
        :rtype: str
        """
        return "Animate2(" + ", ".join([x.type for x in self.displaylist]) + ")"

    def __str__(self):
        return f"Animate2(len={len(self.displaylist)}"

    def artists(self):
        """
        List of artists that need to be updated

        :param self: the animation
        :type self: Animate
        :returns: list of artists
        :rtype: list
        """
        return [x.h for x in self.displaylist]

    def _draw(self, T):
        for x in self.displaylist:
            x.draw(T)

    # ------------------- plot()

    def set_aspect(self, *args, **kwargs):
        self.ax.set_aspect(*args, **kwargs)

    def autoscale(self, *args, **kwargs):
        #self.ax.autoscale(*args, **kwargs)
        pass

    class _Line:
        def __init__(self, anim, h, xs, ys):
            # form 3xN matrix, columns are first/last point in homogeneous form
            p = np.vstack([xs, ys])
            self.p = np.vstack([p, np.ones((p.shape[1],))])
            self.h = h
            self.type = "line"
            self.anim = anim

        def draw(self, T):
            p = T @ self.p
            self.h.set_data(p[0, :], p[1, :])

    def plot(self, x, y, *args, **kwargs):
        """
        Plot a polyline

        :param x: list of x-coordinates
        :type x: array_like
        :param y: list of y-coordinates
        :type y: array_like

        Other arguments as accepted by the matplotlib method.

        All arrays must have the same length.

        :seealso: :func:`matplotlib.pyplot.plot`
        """

        (h,) = self.ax.plot(x, y, *args, **kwargs)
        self.displaylist.append(Animate2._Line(self, h, x, y))
        return h

    # ------------------- quiver()

    class _Quiver:
        def __init__(self, anim, h, x, y, u, v):
            self.type = "quiver"
            self.anim = anim

            self.h = h
            self.type = "arrow"
            self.anim = anim

            self.p = np.c_[u - x, v - y].T

        def draw(self, T):
            R, t = base.tr2rt(T)
            p = R @ self.p
            # specific to a single Quiver
            self.h.set_offsets(t)  # shift the origin
            self.h.set_UVC(p[0], p[1])

    def quiver(self, x, y, u, v, *args, **kwargs):
        """
        Plot a quiver

        :param x: list of base x-coordinates
        :type x: array_like
        :param y: list of base y-coordinates
        :type y: array_like
        :param u: list of vector x-coordinates
        :type u: array_like
        :param v: list of vector y-coordinates
        :type v: array_like


        Draws a series of arrows, the bases defined by corresponding elements
        of (x,y,z) and the vector has components defined by corresponding
        elements of (u,v,w).

        Other arguments as accepted by the matplotlib method.

        :seealso: :func:`matplotlib.pyplot.quiver`
        """
        h = self.ax.quiver(x, y, u, v, *args, **kwargs)
        self.displaylist.append(Animate2._Quiver(self, h, x, y, u, v))

    # ------------------- text()

    class _Text:
        def __init__(self, anim, h, x, y):
            self.type = "text"
            self.h = h
            self.p = np.r_[x, y, 1]
            self.anim = anim

        def draw(self, T):
            p = T @ self.p
            # x2, y2, _ = proj3d.proj_transform(
            #   p[0], p[1], p[2], self.anim.ax.get_proj())
            # self.h.set_position((x2, y2))
            self.h.set_position((p[0], p[1]))

    def text(self, x, y, *args, **kwargs):
        """
        Plot text

        :param x: x-coordinate
        :type x: float
        :param y: float
        :type y: float
        :param z: z-coordinate
        :type z: float
        :param kwargs: Other arguments as accepted by the matplotlib method.

        ``.text(x, y, s)`` display the string ``s`` at coordinate
        (``x``, ``y``).

        :seealso: :func:`matplotlib.pyplot.text`
        """
        h = self.ax.text(x, y, *args, **kwargs)
        self.displaylist.append(Animate2._Text(self, h, x, y))

    # ------------------- scatter()

    def scatter(self, x, y, s=0, **kwargs):
        h = self.plot(x, y, '.', markersize=0, **kwargs)
        self.displaylist.append(Animate2._Line(self, h, x, y))

    # ------------------- wrappers for Axes primitives

    def set_xlim(self, *args, **kwargs):
        self.ax.set_xlim(*args, **kwargs)

    def set_ylim(self, *args, **kwargs):
        self.ax.set_ylim(*args, **kwargs)

    def set_xlabel(self, *args, **kwargs):
        self.ax.set_xlabel(*args, **kwargs)

    def set_ylabel(self, *args, **kwargs):
        self.ax.set_ylabel(*args, **kwargs)


if __name__ == "__main__":
    # from spatialmath import UnitQuaternion
    # from spatialmath.base import tranimate, r2t

    # J = np.array([[2, -1, 0], [-1, 4, 0], [0, 0, 3]])
    # dt = 0.05
    # def attitude():
    #     attitude = UnitQuaternion()
    #     w = 0.2 * np.r_[1, 2, 2].T
    #     for t in np.arange(0, 3, dt):
    #         wd =  -np.linalg.inv(J) @ (np.cross(w, J @ w))
    #         w += wd * dt
    #         attitude.increment(w * dt)
    #         yield attitude.R
    # plt.figure()
    # plotvol3(2)
    # tranimate(attitude())

    from spatialmath import base

    # T = base.rpy2r(0.3, 0.4, 0.5)
    # base.tranimate(T, wait=True)

    T = base.rot2(2)
    base.tranimate2(T, wait=True)



# Part of Spatial Math Toolbox for Python
# Copyright (c) 2000 Peter Corke
# MIT Licence, see details in top-level file: LICENCE

"""
This modules contains functions to operate on special matrices in 2D or 3D, for
example SE(n), SO(n), se(n) and so(n) where n is 2 or 3.

Vector arguments are what numpy refers to as ``array_like`` and can be a list,
tuple, numpy array, numpy row vector or numpy column vector.
"""
# pylint: disable=invalid-name

import math
import numpy as np
from spatialmath import base

# from spatialmath.base import vectors as vec
# from spatialmath.base import transforms2d as t2d
# from spatialmath.base import transforms3d as t3d
# from spatialmath.base import argcheck
# from spatialmath.base import symbolic as sym

try:  # pragma: no cover
    # print('Using SymPy')
    from sympy import Matrix

    _symbolics = True

except ImportError:  # pragma: no cover
    _symbolics = False

_eps = np.finfo(np.float64).eps

# ---------------------------------------------------------------------------------------#
def r2t(R, check=False):
    """
    Convert SO(n) to SE(n)

    :param R: rotation matrix
    :type R: ndarray(2,2) or ndarray(3,3)
    :param check: check if rotation matrix is valid (default False, no check)
    :type check: bool
    :return: homogeneous transformation matrix
    :rtype: ndarray(3,3) or ndarray(4,4)
    :raises ValueError: bad argument

    ``T = r2t(R)`` is an SE(2) or SE(3) homogeneous transform equivalent to an
    SO(2) or SO(3) orthonormal rotation matrix ``R`` with a zero translational
    component

    - if ``R`` is 2x2 then ``T`` is 3x3: SO(2) → SE(2)
    - if ``R`` is 3x3 then ``T`` is 4x4: SO(3) → SE(3)

    .. runblock:: pycon

        >>> from spatialmath.base import *
        >>> R = rot2(0.3)
        >>> R
        >>> r2t(R)

    :seealso: t2r, rt2tr
    """
    if not isinstance(R, np.ndarray):
        raise ValueError("argument must be NumPy array")
    dim = R.shape
    if dim[0] != dim[1]:
        raise ValueError("Matrix must be square")
    n = dim[0] + 1
    m = dim[0]

    if R.dtype == "O":
        # symbolic matrix
        T = np.zeros((n, n), dtype="O")
    else:
        # numeric matrix
        if not isinstance(R, np.ndarray):
            raise ValueError("Argument must be a NumPy array")
        if check and not isR(R):
            raise ValueError("Invalid SO(3) matrix ")

        # T = np.pad(R, (0, 1), mode='constant')
        # T[-1, -1] = 1.0
        T = np.zeros((n, n))
    T[:m, :m] = R
    T[-1, -1] = 1

    return T


# ---------------------------------------------------------------------------------------#
def t2r(T, check=False):
    """
    Convert SE(n) to SO(n)

    :param T: homogeneous transformation matrix
    :type T: ndarray(3,3) or ndarray(4,4)
    :param check: check if rotation matrix is valid (default False, no check)
    :type check: bool
    :return: rotation matrix
    :rtype: ndarray(2,2) or ndarray(3,3)
    :raises ValueError: bad argument

    ``R = T2R(T)`` is the orthonormal rotation matrix component of homogeneous
    transformation matrix ``T``

    - if ``T`` is 3x3 then ``R`` is 2x2: SE(2) → SO(2)
    - if ``T`` is 4x4 then ``R`` is 3x3: SE(3) → SO(3)

    .. runblock:: pycon

        >>> from spatialmath.base import *
        >>> T = trot2(0.3, t=[1,2])
        >>> T
        >>> t2r(T)

    .. note:: Any translational component of T is lost.

    :seealso: r2t, tr2rt
    """
    if not isinstance(T, np.ndarray):
        raise ValueError("argument must be NumPy array")
    dim = T.shape
    if dim[0] != dim[1]:
        raise ValueError("Matrix must be square")

    if dim[0] == 3:
        R = T[:2, :2]
    elif dim[0] == 4:
        R = T[:3, :3]
    else:
        raise ValueError("Value must be an SE(3) matrix")

    if check and not isR(R):
        raise ValueError("Invalid rotation submatrix")

    return R


# ---------------------------------------------------------------------------------------#


def tr2rt(T, check=False):
    """
    Convert SE(n) to SO(n) and translation

    :param T: SE(n) matrix
    :type T: ndarray(3,3) or ndarray(4,4)
    :param check: check if SO(3) submatrix is valid (default False, no check)
    :type check: bool
    :return: SO(n) matrix and translation vector
    :rtype: tuple: (ndarray(2,2), ndarray(2)) or (ndarray(3,3), ndarray(3))
    :raises ValueError: bad argument

    (R,t) = tr2rt(T) splits a homogeneous transformation matrix (NxN) into an orthonormal
    rotation matrix R (MxM) and a translation vector T (Mx1), where N=M+1.

    - if ``T`` is 3x3 - in SE(2) - then ``R`` is 2x2 and ``t`` is 2x1.
    - if ``T`` is 4x4 - in SE(3) - then ``R`` is 3x3 and ``t`` is 3x1.

    .. runblock:: pycon

        >>> from spatialmath.base import *
        >>> T = trot2(0.3, t=[1,2])
        >>> T
        >>> R, t = tr2rt(T)
        >>> R
        >>> t

    :seealso: rt2tr, tr2r
    """
    if not isinstance(T, np.ndarray):
        raise ValueError("argument must be NumPy array")
    dim = T.shape
    if dim[0] != dim[1]:
        raise ValueError("Matrix must be square")

    if dim[0] == 3:
        R = t2r(T, check)
        t = T[:2, 2]
    elif dim[0] == 4:
        R = t2r(T, check)
        t = T[:3, 3]
    else:
        raise ValueError("T must be an SE2 or SE3 homogeneous transformation matrix")

    return [R, t]


# ---------------------------------------------------------------------------------------#


def rt2tr(R, t, check=False):
    """
    Convert SO(n) and translation to SE(n)

    :param R: SO(n) matrix
    :type R: ndarray(2,2) or ndarray(3,3)
    :param t: translation vector
    :type R: ndarray(2) or ndarray(3)
    :param check: check if SO(3) matrix is valid (default False, no check)
    :type check: bool
    :return: SE(3) matrix
    :rtype: ndarray(4,4) or (3,3)
    :raises ValueError: bad argument

    ``T = rt2tr(R, t)`` is a homogeneous transformation matrix (N+1xN+1) formed from an
    orthonormal rotation matrix ``R`` (NxN) and a translation vector ``t``
    (Nx1).

    - If ``R`` is 2x2 and ``t`` is 2x1, then ``T`` is 3x3
    - If ``R`` is 3x3 and ``t`` is 3x1, then ``T`` is 4x4

    .. runblock:: pycon

        >>> from spatialmath.base import *
        >>> R = rot2(0.3)
        >>> t = [1, 2]
        >>> rt2tr(R, t)

    :seealso: rt2m, tr2rt, r2t
    """
    t = base.getvector(t, dim=None, out="array")
    if not isinstance(R, np.ndarray):
        raise ValueError("Rotation matrix not a NumPy array")
    if R.shape[0] != t.shape[0]:
        raise ValueError("R and t must have the same number of rows")
    if check and not isR(R):
        raise ValueError("Invalid rotation matrix")

    if R.dtype == "O":
        if R.shape == (2, 2):
            T = np.pad(R, ((0, 1), (0, 1)), 'constant')
            T[:2, 2] = t
            T[2, 2] = 1
        elif R.shape == (3, 3):
            T = np.pad(R, ((0, 1), (0, 1)), 'constant')
            T[:3, 3] = t
            T[3, 3] = 1
        else:
            raise ValueError("R must be an SO2 or SO3 rotation matrix")
    else:

        if R.shape == (2, 2):
            T = np.eye(3)
            T[:2, :2] = R
            T[:2, 2] = t
        elif R.shape == (3, 3):
            T = np.eye(4)
            T[:3, :3] = R
            T[:3, 3] = t
        else:
            raise ValueError("R must be an SO2 or SO3 rotation matrix")

    return T


# ---------------------------------------------------------------------------------------#


def Ab2M(A, b):
    """
    Pack matrix and vector to matrix

    :param A: square matrix
    :type A: ndarray(3,3) or ndarray(2,2)
    :param b: translation vector
    :type b: ndarray(3) or ndarray(2)
    :return: matrix
    :rtype: ndarray(4,4) or ndarray(3,3)
    :raises ValueError: bad arguments

    ``M = Ab2M(A, b)`` is a matrix (N+1xN+1) formed from a matrix ``R`` (NxN) and a vector ``t``
    (Nx1).  The bottom row is all zeros.

    - If ``A`` is 2x2 and ``b`` is 2x1, then ``M`` is 3x3
    - If ``A`` is 3x3 and ``b`` is 3x1, then ``M`` is 4x4

    .. runblock:: pycon

        >>> from spatialmath.base import *
        >>> A = np.c_[[1, 2], [3, 4]].T
        >>> b = [5, 6]
        >>> Ab2M(A, b)

    :seealso: rt2tr, tr2rt, r2t
    """
    b = base.getvector(b, dim=None, out="array")
    if not isinstance(A, np.ndarray):
        raise ValueError("Rotation matrix not a NumPy array")
    if A.shape[0] != b.shape[0]:
        raise ValueError("A and b must have the same number of rows")

    if A.shape == (2, 2):
        T = np.zeros((3, 3))
        T[:2, :2] = A
        T[:2, 2] = b
    elif A.shape == (3, 3):
        T = np.zeros((4, 4))
        T[:3, :3] = A
        T[:3, 3] = b
    else:
        raise ValueError("A must be 2x2 or 3x3")

    return T


# ======================= predicates


def isR(R, tol=100):
    r"""
    Test if matrix belongs to SO(n)

    :param R: matrix to test
    :type R: ndarray(2,2) or ndarray(3,3)
    :param tol: tolerance in units of eps
    :type tol: float
    :return: whether matrix is a proper orthonormal rotation matrix
    :rtype: bool

    Checks orthogonality, ie. :math:`{\bf R} {\bf R}^T = {\bf I}` and :math:`\det({\bf R}) > 0`.
    For the first test we check that the norm of the residual is less than ``tol * eps``.

    .. runblock:: pycon

        >>> from spatialmath.base import *
        >>> isR(np.eye(3))
        >>> isR(rot2(0.5))
        >>> isR(np.zeros((3,3)))

    :seealso: isrot2, isrot
    """
    return (
        np.linalg.norm(R @ R.T - np.eye(R.shape[0])) < tol * _eps
        and np.linalg.det(R @ R.T) > 0
    )


def isskew(S, tol=10):
    r"""
    Test if matrix belongs to so(n)

    :param S: matrix to test
    :type S: ndarray(2,2) or ndarray(3,3)
    :param tol: tolerance in units of eps
    :type tol: float
    :return: whether matrix is a proper skew-symmetric matrix
    :rtype: bool

    Checks skew-symmetry, ie. :math:`{\bf S} + {\bf S}^T = {\bf 0}`.
    We check that the norm of the residual is less than ``tol * eps``.

    .. runblock:: pycon

        >>> from spatialmath.base import *
        >>> import numpy as np
        >>> isskew(np.zeros((3,3)))
        >>> isskew(np.array([[0, -2], [2, 0]]))
        >>> isskew(np.eye(3))

    :seealso: isskewa
    """
    return np.linalg.norm(S + S.T) < tol * _eps


def isskewa(S, tol=10):
    r"""
    Test if matrix belongs to se(n)

    :param S: matrix to test
    :type S: ndarray(3,3) or ndarray(4,4)
    :param tol: tolerance in units of eps
    :type tol: float
    :return: whether matrix is a proper skew-symmetric matrix
    :rtype: bool

    Check if matrix is augmented skew-symmetric, ie. the top left (n-1xn-1) partition ``S`` is
    skew-symmetric :math:`{\bf S} + {\bf S}^T = {\bf 0}`, and the bottom row is zero
    We check that the norm of the residual is less than ``tol * eps``.

    .. runblock:: pycon

        >>> from spatialmath.base import *
        >>> import numpy as np
        >>> isskewa(np.zeros((3,3)))
        >>> isskewa(np.array([[0, -2], [2, 0]])) # this matrix is skew but not skewa
        >>> isskewa(np.array([[0, -2, 5], [2, 0, 6], [0, 0, 0]]))

    :seealso: isskew
    """
    return np.linalg.norm(S[0:-1, 0:-1] + S[0:-1, 0:-1].T) < tol * _eps and np.all(
        S[-1, :] == 0
    )


def iseye(S, tol=10):
    """
    Test if matrix is identity

    :param S: matrix to test
    :type S: ndarray(n,n)
    :param tol: tolerance in units of eps
    :type tol: float
    :return: whether matrix is a proper skew-symmetric matrix
    :rtype: bool

    Check if matrix is an identity matrix. We test that the trace tom row is zero
    We check that the norm of the residual is less than ``tol * eps``.

    .. runblock:: pycon

        >>> from spatialmath.base import *
        >>> import numpy as np
        >>> iseye(np.array([[1,0], [0,1]]))
        >>> iseye(np.array([[1,2], [0,1]]))

    :seealso: isskew, isskewa
    """
    s = S.shape
    if len(s) != 2 or s[0] != s[1]:
        return False  # not a square matrix
    return np.linalg.norm(S - np.eye(s[0])) < tol * _eps


# ---------------------------------------------------------------------------------------#
def skew(v):
    r"""
    Create skew-symmetric metrix from vector

    :param v: vector
    :type v: array_like(1) or array_like(3)
    :return: skew-symmetric matrix in so(2) or so(3)
    :rtype: ndarray(2,2) or ndarray(3,3)
    :raises ValueError: bad argument

    ``skew(V)`` is a skew-symmetric matrix formed from the elements of ``V``.

    - ``len(V)``  is 1 then ``S`` = :math:`\left[ \begin{array}{cc} 0 & -v \\ v & 0 \end{array} \right]`
    - ``len(V)`` is 3 then ``S`` = :math:`\left[ \begin{array}{ccc} 0 & -v_z & v_y \\ v_z & 0 & -v_x \\ -v_y & v_x & 0\end{array} \right]`

    .. runblock:: pycon

        >>> from spatialmath.base import *
        >>> skew(2)
        >>> skew([1, 2, 3])

    .. note::

        - This is the inverse of the function ``vex()``.
        - These are the generator matrices for the Lie algebras so(2) and so(3).

    :seealso: :func:`vex`, :func:`skewa`
    :SymPy: supported
    """
    v = base.getvector(v, None, "sequence")
    if len(v) == 1:
        return np.array([[0, -v[0]], [v[0], 0]])
    elif len(v) == 3:
        return np.array([[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]])
    else:
        raise ValueError("argument must be a 1- or 3-vector")


# ---------------------------------------------------------------------------------------#


def vex(s, check=False):
    r"""
    Convert skew-symmetric matrix to vector

    :param s: skew-symmetric matrix
    :type s: ndarray(2,2) or ndarray(3,3)
    :param check: check if matrix is skew symmetric (default False, no check)
    :type check: bool
    :return: vector of unique values
    :rtype: ndarray(1) or ndarray(3)
    :raises ValueError: bad argument

    ``vex(S)`` is the vector which has the corresponding skew-symmetric matrix ``S``.

    - ``S`` is 2x2 - so(2) case - where ``S`` :math:`= \left[ \begin{array}{cc} 0 & -v \\ v & 0 \end{array} \right]` then return :math:`[v]`
    - ``S`` is 3x3 - so(3) case -  where ``S`` :math:`= \left[ \begin{array}{ccc} 0 & -v_z & v_y \\ v_z & 0 & -v_x \\ -v_y & v_x & 0\end{array} \right]` then return :math:`[v_x, v_y, v_z]`.

    .. runblock:: pycon

        >>> from spatialmath.base import *
        >>> S = skew(2)
        >>> print(S)
        >>> vex(S)
        >>> S = skew([1, 2, 3])
        >>> print(S)
        >>> vex(S)

    .. note::

        - This is the inverse of the function ``skew()``.
        - Only rudimentary checking (zero diagonal) is done to ensure that the matrix
          is actually skew-symmetric.
        - The function takes the mean of the two elements that correspond to each unique
          element of the matrix.

    :seealso: :func:`skew`, :func:`vexa`
    :SymPy: supported
    """
    if s.shape == (3, 3):
        if check and not isskew(s):
            raise ValueError("Argument is not skew symmetric")
        return np.array([s[2, 1] - s[1, 2], s[0, 2] - s[2, 0], s[1, 0] - s[0, 1]]) / 2
    elif s.shape == (2, 2):
        return np.array([s[1, 0] - s[0, 1]]) / 2
    else:
        raise ValueError("Argument must be 2x2 or 3x3 matrix")


# ---------------------------------------------------------------------------------------#


def skewa(v):
    r"""
    Create augmented skew-symmetric metrix from vector

    :param v: vector
    :type v: array_like(3), array_like(6)
    :return: augmented skew-symmetric matrix in se(2) or se(3)
    :rtype: ndarray(3,3) or ndarray(4,4)
    :raises ValueError: bad argument

    ``skewa(V)`` is an augmented skew-symmetric matrix formed from the elements of ``V``.

    - ``len(V)`` is 3 then S = :math:`\left[ \begin{array}{ccc} 0 & -v_3 & v_1 \\ v_3 & 0 & v_2 \\ 0 & 0 & 0 \end{array} \right]`
    - ``len(V)`` is 6 then S = :math:`\left[ \begin{array}{cccc} 0 & -v_6 & v_5 & v_1 \\ v_6 & 0 & -v_4 & v_2 \\ -v_5 & v_4 & 0 & v_3 \\ 0 & 0 & 0 & 0 \end{array} \right]`

    .. runblock:: pycon

        >>> from spatialmath.base import *
        >>> skewa([1, 2, 3])
        >>> skewa([1, 2, 3, 4, 5, 6])

    .. note::

        - This is the inverse of the function ``vexa()``.
        - These are the generator matrices for the Lie algebras se(2) and se(3).
        - Map twist vectors in 2D and 3D space to se(2) and se(3).

    :seealso: :func:`vexa`, :func:`skew`
    :SymPy: supported
    """

    v = base.getvector(v, None)
    if len(v) == 3:
        omega = np.zeros((3, 3), dtype=v.dtype)
        omega[:2, :2] = skew(v[2])
        omega[:2, 2] = v[0:2]
        return omega
    elif len(v) == 6:
        omega = np.zeros((4, 4), dtype=v.dtype)
        omega[:3, :3] = skew(v[3:6])
        omega[:3, 3] = v[0:3]
        return omega
    else:
        raise ValueError("expecting a 3- or 6-vector")


def vexa(Omega, check=False):
    r"""
    Convert skew-symmetric matrix to vector

    :param s: augmented skew-symmetric matrix
    :type s: ndarray(3,3) or ndarray(4,4)
    :param check: check if matrix is skew symmetric part is valid (default False, no check)
    :type check: bool
    :return: vector of unique values
    :rtype: ndarray(3) or ndarray(6)
    :raises ValueError: bad argument

    ``vexa(S)`` is the vector which has the corresponding augmented skew-symmetric matrix ``S``.

    - ``S`` is 3x3 - se(2) case - where ``S`` :math:`= \left[ \begin{array}{ccc} 0 & -v_3 & v_1 \\ v_3 & 0 & v_2 \\ 0 & 0 & 0 \end{array} \right]` then return :math:`[v_1, v_2, v_3]`.
    - ``S`` is 4x4 - se(3) case -  where ``S`` :math:`= \left[ \begin{array}{cccc} 0 & -v_6 & v_5 & v_1 \\ v_6 & 0 & -v_4 & v_2 \\ -v_5 & v_4 & 0 & v_3 \\ 0 & 0 & 0 & 0 \end{array} \right]` then return :math:`[v_1, v_2, v_3, v_4, v_5, v_6]`.

    .. runblock:: pycon

        >>> from spatialmath.base import *
        >>> S = skewa([1, 2, 3])
        >>> print(S)
        >>> vexa(S)
        >>> S = skewa([1, 2, 3, 4, 5, 6])
        >>> print(S)
        >>> vexa(S)

    .. note::

        - This is the inverse of the function ``skewa``.
        - Only rudimentary checking (zero diagonal) is done to ensure that the matrix
          is actually skew-symmetric.
        - The function takes the mean of the two elements that correspond to each unique
          element of the matrix.

    :seealso: :func:`skewa`, :func:`vex`
    :SymPy: supported
    """
    if Omega.shape == (4, 4):
        return np.hstack((base.transl(Omega), vex(t2r(Omega), check=check)))
    elif Omega.shape == (3, 3):
        return np.hstack((base.transl2(Omega), vex(t2r(Omega), check=check)))
    else:
        raise ValueError("expecting a 3x3 or 4x4 matrix")


def rodrigues(w, theta=None):
    r"""
    Rodrigues' formula for rotation

    :param w: rotation vector
    :type w: array_like(3) or array_like(1)
    :param θ: rotation angle
    :type θ: float or None
    :return: SO(n) matrix
    :rtype: ndarray(2,2) or ndarray(3,3)

    Compute Rodrigues' formula for a rotation matrix given a rotation axis
    and angle.

    .. math::

        \mat{R} = \mat{I}_{3 \times 3} + \sin \theta \skx{\hat{\vec{v}}} + (1 - \cos \theta) \skx{\hat{\vec{v}}}^2

    .. runblock:: pycon

        >>> from spatialmath.base import *
        >>> rodrigues([1, 0, 0], 0.3)
        >>> rodrigues([0.3, 0, 0])
        >>> rodrigues(0.3)   # 2D version

    """
    w = base.getvector(w)
    if base.iszerovec(w):
        # for a zero so(n) return unit matrix, theta not relevant
        if len(w) == 1:
            return np.eye(2)
        else:
            return np.eye(3)
    if theta is None:
        w, theta = base.unitvec_norm(w)

    skw = skew(w)
    return (
        np.eye(skw.shape[0])
        + math.sin(theta) * skw
        + (1.0 - math.cos(theta)) * skw @ skw
    )


def h2e(v):
    """
    Convert from homogeneous to Euclidean form

    :param v: homogeneous vector or matrix
    :type v: array_like(n), ndarray(n,m)
    :return: Euclidean vector
    :rtype: ndarray(n-1), ndarray(n-1,m)

    - If ``v`` is an N-vector, return an (N-1)-column vector where the elements have
      all been scaled by the last element of ``v``.
    - If ``v`` is a matrix (NxM), return a matrix (N-1xM), where each column has
      been scaled by its last element.

    .. runblock:: pycon

        >>> from spatialmath.base import *
        >>> h2e([2, 4, 6, 1])
        >>> h2e([2, 4, 6, 2])
        >>> h = np.c_[[1,2,1], [3,4,2], [5,6,1]]
        >>> h
        >>> h2e(h)

    .. note:: The result is always a 2D array, a 1D input results in a column vector.

    :seealso: e2h
    """
    if isinstance(v, np.ndarray) and len(v.shape) == 2:
        # dealing with matrix
        return v[:-1, :] / v[-1, :][np.newaxis, :]

    elif base.isvector(v):
        # dealing with shape (N,) array
        v = base.getvector(v, out="col")
        return v[0:-1] / v[-1]


def e2h(v):
    """
    Convert from Euclidean to homogeneous form

    :param v: Euclidean vector or matrix
    :type v: array_like(n), ndarray(n,m)
    :return: homogeneous vector
    :rtype: ndarray(n+1,m)

    - If ``v`` is an N-vector, return an (N+1)-column vector where a value of 1 has
      been appended as the last element.
    - If ``v`` is a matrix (NxM), return a matrix (N+1xM), where each column has
      been appended with a value of 1, ie. a row of ones has been appended to the matrix.

    .. runblock:: pycon

        >>> from spatialmath.base import *
        >>> e2h([2, 4, 6])
        >>> e = np.c_[[1,2], [3,4], [5,6]]
        >>> e
        >>> e2h(e)

    .. note:: The result is always a 2D array, a 1D input results in a column vector.

    :seealso: e2h
    """
    if isinstance(v, np.ndarray) and len(v.shape) == 2:
        # dealing with matrix
        return np.vstack([v, np.ones((1, v.shape[1]))])

    elif base.isvector(v):
        # dealing with shape (N,) array
        v = base.getvector(v, out="col")
        return np.vstack((v, 1))


def homtrans(T, p):
    r"""
    Apply a homogeneous transformation to a Euclidean vector

    :param T: homogeneous transformation
    :type T: Numpy array (n,n)
    :param p: Vector(s) to be transformed
    :type p: array_like(n-1), ndarray(n-1,m)
    :return: transformed Euclidean vector(s)
    :rtype: ndarray(n-1,m)
    :raises ValueError: bad argument

    - ``homtrans(T, p)`` applies the homogeneous transformation ``T`` to the Euclidean points
      stored columnwise in the array ``p``.

    - ``homtrans(T, v)`` as above but ``v`` is a 1D array considered to be a column vector, and the
      retured value will be a column vector.


    .. runblock:: pycon

        >>> from spatialmath.base import *
        >>> T = trotx(0.3)
        >>> v = [1, 2, 3]
        >>> h2e( T @ e2h(v))
        >>> homtrans(T, v)

    .. note::

        - If T is a homogeneous transformation defining the pose of {B} with respect to {A},
          then the points are defined with respect to frame {B} and are transformed to be
          with respect to frame {A}.

    :seealso: :func:`e2h`, :func:`h2e`
    """
    p = e2h(p)
    if p.shape[0] != T.shape[0]:
        raise ValueError("matrices and point data do not conform")

    return h2e(T @ p)


def det(m):
    """
    Determinant of matrix

    :param m: any square matrix
    :type v: array_like(n,n)
    :return: determinant
    :rtype: float

    ``det(v)`` is the determinant of the matrix ``m``.

    .. runblock:: pycon

        >>> from spatialmath.base import *
        >>> norm([3, 4])

    :seealso: :func:`~numpy.linalg.det`

    :SymPy: supported
    """
    if m.dtype.kind == "O":
        return Matrix(m).det()
    else:
        return np.linalg.det(m)


if __name__ == "__main__":  # pragma: no cover
    import pathlib

    print(e2h((1, 2, 3)))
    print(h2e((1, 2, 3)))
    exec(
        open(
            pathlib.Path(__file__).parent.absolute() / "test" / "test_transformsNd.py"
        ).read()
    )  # pylint: disable=exec-used
